// View Synthesis Decoder
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.

#include "ViewSynthesis.h"
#include <string.h>
#include <stdlib.h>
#include <opencv2/legacy/compat.hpp> // for cvmMul()
#include <opencv2/calib3d/calib3d.hpp> // for cvFindHomography()
#include <opencv2/imgproc/imgproc_c.h> // for cvSmooth(), cvDilate(), cvInpaint()
#include <opencv2/photo/photo_c.h> // for cvInpaint()
#include <opencv2/highgui/highgui_c.h> // for DEBUG cvSaveImage()
#include <opencv2/imgproc/types_c.h> // for DEBUG CV_YUV2RGB

#ifndef SAFE_RELEASE_IMAGE
#define SAFE_RELEASE_IMAGE(p) { if((p)!=NULL){ cvReleaseImage(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_RELEASE_MAT
#define SAFE_RELEASE_MAT(p) { if((p)!=NULL){ cvReleaseMat(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_FREE
#define SAFE_FREE(p) { if((p)!=NULL){ free(p); (p)=NULL; } }
#endif

using namespace std;

CViewInterpolation::CViewInterpolation()
{
  int i;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::CViewInterpolation()\n");
  #endif

  m_uiPrecision= 1; // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

  m_imgVirtualImage= m_imgVirtualDepth= m_imgWorkDepth= NULL;
  m_imgSuccessSynthesis = m_imgHoles = NULL;
  m_pVirtualDepth= m_pSuccessSynthesis= m_pWorkDepth= NULL;

  m_pS= m_pV= m_pR= m_pD= NULL;

  for(i=0; i<256; i++)
    m_matH_R2V[i] = m_matH_V2R[i] = NULL;

  // for VariParam
  for(i=0; i<2; i++)
    m_Mat_in[i]= m_Mat_ex_c2w[i]= m_Mat_proj_w2i[i]= NULL;

  cv::setUseOptimized(true);
}



CViewInterpolation::~CViewInterpolation()
{
  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::~CViewInterpolation()\n");
  #endif

  xReleaseMemory();
}



void CViewInterpolation::xReleaseMemory()
{
  int i;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::xReleaseMemory()\n");
  #endif

  SAFE_RELEASE_IMAGE(m_imgVirtualImage);
  SAFE_RELEASE_IMAGE(m_imgVirtualDepth);
  SAFE_RELEASE_IMAGE(m_imgWorkDepth);
  SAFE_RELEASE_IMAGE(m_imgSuccessSynthesis);
  SAFE_RELEASE_IMAGE(m_imgHoles);

  if(m_pS!=NULL) delete m_pS; m_pS= NULL;
  if(m_pV!=NULL) delete m_pV; m_pV= NULL;
  if(m_pR!=NULL) delete m_pR; m_pR= NULL;
  if(m_pD!=NULL) delete m_pD; m_pD= NULL;
  // if(m_pDB!=NULL) delete m_pDB; m_pDB= NULL;

  SAFE_FREE(m_pVirtualDepth);
  SAFE_FREE(m_pWorkDepth);
  SAFE_FREE(m_pSuccessSynthesis);

  for (i=0; i<256; i++) {
    SAFE_RELEASE_MAT( m_matH_R2V[i] );
    SAFE_RELEASE_MAT( m_matH_V2R[i] );
  }

  // for VariParam
  for(i=0; i<2; i++) {
    SAFE_RELEASE_MAT( m_Mat_in[i] );
    SAFE_RELEASE_MAT( m_Mat_ex_c2w[i] );
    SAFE_RELEASE_MAT( m_Mat_proj_w2i[i] );
  }
}



void CViewInterpolation::convertCameraParam(CvMat *exMat_dst, CvMat *exMat_src)
{
  int i, j;
  double val;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::convertCameraParam()\n");
  #endif

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      cvmSet(exMat_dst, i, j, cvmGet(exMat_src, j, i));

  for (i=0; i<3; i++) {
    val = 0.0;
    for(j=0; j<3; j++)
      val -= cvmGet(exMat_dst, i, j)*cvmGet(exMat_src, j, 3);
    cvmSet(exMat_dst, i, 3, val);
  }
}



// Winter Versionでは、Call側でカメラ位置を左右反転させて指定する。
// LEFT:  C1, C2, V1, T1
// RIGHT: C3, C2, V2, T2
bool CViewInterpolation::InitOneSet (
    unsigned int width, unsigned int height, bool variparam,
    double near1, double far1, double near2, double far2,
    const char *RefCam1, const char *RefCam2, const char *VirCam, const char *TarCam,
    const unsigned int Side,
    double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
    double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
    double Mat_In_V[9], double Mat_Ex_V[9], double Mat_Trans_V[3],
    double Mat_In_T[9], double Mat_Ex_T[9], double Mat_Trans_T[3]
) {
  VariParam= variparam;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::InitOneSet()\n");
  #endif

  m_ResidualsSide= Side;
  m_uiWidth= width;
  m_uiHeight= height;

  /* スタックファイル直接読み込み用
  m_pDB= new CViewInterpolation();
  if (!m_pDB->InitOneCh(
              width/2, height/2, variparam,
              near1, far1,
              VirCam, TarCam, Side,
              Mat_In_V, Mat_Ex_V, Mat_Trans_V,
              Mat_In_T, Mat_Ex_T, Mat_Trans_T
  ) ) return false;
  */

  m_pD= new CViewInterpolation();
  if (!m_pD->InitOneCh(
              width, height, variparam,
              near1, far1,
              VirCam, TarCam, Side,
              Mat_In_V, Mat_Ex_V, Mat_Trans_V,
              Mat_In_T, Mat_Ex_T, Mat_Trans_T
  ) ) return false;

  //// SHIFT ch /////////////////////////////////////////////////////////
  m_pS= new CViewInterpolation();
  if (!m_pS->InitOneCh(
              width, height, variparam,
              near1, far1,
              VirCam, TarCam, Side,
              Mat_In_V, Mat_Ex_V, Mat_Trans_V,
              Mat_In_T, Mat_Ex_T, Mat_Trans_T
  ) ) return false;

  //// SUMMER: RIGHT ch, WINTER: View ch ////////////////////////////////////////////////////////
  m_pV= new CViewInterpolation();
  if (!m_pV->InitOneCh(
              width, height, variparam,
              near1, far1,
              RefCam2, TarCam, Side,
              Mat_In_2, Mat_Ex_2, Mat_Trans_2,
              Mat_In_T, Mat_Ex_T, Mat_Trans_T
  ) ) return false;

  //// SUMMER: Left ch, WINTER: Residuals ch  ////////////////////////////////////////////////////////
  m_pR= new CViewInterpolation();
  if (!m_pR->InitOneCh(
              width, height, variparam,
              near2, far2,
              RefCam1, TarCam, Side,
              Mat_In_1, Mat_Ex_1, Mat_Trans_1,
              Mat_In_T, Mat_Ex_T, Mat_Trans_T
  ) ) return false;

  m_pV->m_weightVR= fabs(Mat_Trans_1[0] - Mat_Trans_T[0]) / fabs(Mat_Trans_2[0] - Mat_Trans_T[0]);

  return true;
}



bool CViewInterpolation::InitOneCh(
              unsigned int width, unsigned int height, bool variparam,  // double disp255,
              double dNear, double dFar,
              const char *RefCamID, const char *VirCamID, unsigned int Side,
              double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
              double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
) {
  unsigned int h, pos;
  double fB;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::InitOneCh()\n");
  #endif

  xReleaseMemory();

  m_uiPrecision=1; //  4;  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  m_uiResidualsWidth= WIDE;
  m_ResidualsSide= Side;
  m_uiHeight= height;
  m_uiWidth= width;
  RefCameraName= RefCamID;
  VirtualCameraName= VirCamID;

  fB = fabs(Mat_In_Ref[0]) * fabs(Mat_Trans_Ref[0] - Mat_Trans_Vir[0]);
  m_dMaxDisp = fB / fabs(dNear); //  - fB / fabs(dFar);
  m_dispA = 1.0 / 255.0 * fB * fabs(1.0 / dNear - 1.0 / dFar);
  m_dispB = fB / fabs(dFar);

  m_imgVirtualImage = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  m_imgVirtualDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgWorkDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgSuccessSynthesis = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgHoles    = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  if((m_pVirtualDepth = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgVirtualDepth->widthStep)
    m_pVirtualDepth[h] = (BYTE *) &(m_imgVirtualDepth->imageData[pos]);

  if((m_pWorkDepth = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgWorkDepth->widthStep)
    m_pWorkDepth[h] = (BYTE *) &(m_imgWorkDepth->imageData[pos]);

  if((m_pSuccessSynthesis = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgSuccessSynthesis->widthStep)
    m_pSuccessSynthesis[h] = (BYTE *) &(m_imgSuccessSynthesis->imageData[pos]);

//  if (VariParam) {
  for(unsigned int i=0; i<2; i++) {
    m_Mat_in[i] = cvCreateMat(3, 3, CV_64F); //intrinsic parameter of camera (3x3 matrix)
    m_Mat_ex_c2w[i] = cvCreateMat(3, 4, CV_64F); //extrinsic parameter of camera (3x4 matrix)
    m_Mat_proj_w2i[i] = cvCreateMat(3, 4, CV_64F); // projection matrix
  }

  for (unsigned int i=0; i<9; i++) Mat_In_C[i]= Mat_In_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_C[i]= Mat_Ex_Ref[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_C[i]= Mat_Trans_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_In_V[i]= Mat_In_Vir[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_V[i]= Mat_Ex_Vir[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_V[i]= Mat_Trans_Vir[i];
//  }

  return init_3Dwarp( dNear, dFar );

}



bool CViewInterpolation::init_camera_param(
                            double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
                            double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3],
                            CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]
) {
  int i, j;
  CvMat *mat_ex_w2c;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::init_camera_param()\n");
  #endif

  for(i=0; i<3; i++)
  {
    for(j=0; j<3;j++)
    {
      mat_in[0]->data.db[j+3*i] = *(Mat_In_Ref+3*i+j);
      mat_in[1]->data.db[j+3*i] = *(Mat_In_Vir+3*i+j);
    }
  }
  for(i=0; i<3; i++)
  {
    for(j=0; j<3;j++)
    {
      mat_ex_c2w[0]->data.db[j+4*i] = *(Mat_Ex_Ref+3*i+j);
      mat_ex_c2w[1]->data.db[j+4*i] = *(Mat_Ex_Vir+3*i+j);
    }
  }
  for(i=0; i<3; i++)
  {
    mat_ex_c2w[0]->data.db[3+4*i] = *(Mat_Trans_Ref+i);
    mat_ex_c2w[1]->data.db[3+4*i] = *(Mat_Trans_Vir+i);
  }

  if(mat_proj!=NULL)
  {
    mat_ex_w2c = cvCreateMat(3, 4, CV_64F);
    for(i=0; i<2; i++)
    {
      convertCameraParam(mat_ex_w2c, mat_ex_c2w[i]);
      cvmMul(mat_in[i], mat_ex_w2c, mat_proj[i]);  // Proj = inMat_c2i * exMat_w2c
    }
    cvReleaseMat(&mat_ex_w2c);
  }

  return true;
}



void CViewInterpolation::image2world_with_z(
                        CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from, CvMat *image, CvMat *world
) {
  CvMat *temp;
  double s;

  #ifdef PTRACE
    //printf("TRACE: CViewInterpolation::image2world_with_z()\n");
  #endif

  temp= cvCreateMat(3, 1, CV_64F);

  cvmMul(mat_Rc2w_invIN_from, image, temp);

  s= ( cvmGet(world, 2, 0) - cvmGet(matEX_c2w_from, 2, 3) ) / cvmGet(temp, 2, 0);

  cvmSet(world, 0, 0, s*cvmGet(temp, 0, 0) + cvmGet(matEX_c2w_from, 0, 3));
  cvmSet(world, 1, 0, s*cvmGet(temp, 1, 0) + cvmGet(matEX_c2w_from, 1, 3));

  cvReleaseMat(&temp);
}



void CViewInterpolation::makeHomography(
                        CvMat *matH_F2T[256], CvMat *matH_T2F[256],
                        double adTable[256],
                        CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to,
                        unsigned int work
) {
  int i, j, k;
  double val, u, v;
  CvMat *matIN_inv_from, *mat_Rc2w_invIN_from;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::makeHomography()\n");
  #endif

  CvMat *src_points= cvCreateMat(4, 2, CV_64F);
  CvMat *dst_points= cvCreateMat(4, 2, CV_64F);
  CvMat *image= cvCreateMat(3, 1, CV_64F);
  CvMat *world= cvCreateMat(4, 1, CV_64F);

  matIN_inv_from= cvCreateMat(3, 3, CV_64F);
  mat_Rc2w_invIN_from= cvCreateMat(3, 3, CV_64F);

  cvInvert(matIN_from, matIN_inv_from, CV_LU);
  for(i=0; i<3; i++)
  {
    for(j=0; j<3; j++)
    {
      val=0.0;
      for(k=0; k<3; k++)
      {
        val += cvmGet(matEX_c2w_from, i, k)*cvmGet(matIN_inv_from, k, j);
      }
      cvmSet(mat_Rc2w_invIN_from, i, j, val);
    }
  }

  if ( work == INIT )
    for (unsigned int i=0; i<256; i++) {
      matH_F2T[i]= cvCreateMat(3, 3, CV_64F);
      matH_T2F[i]= cvCreateMat(3, 3, CV_64F);
    }

  for(i=0; i<256; i++) {
    cvmSet(world, 2, 0, adTable[i]);
    cvmSet(world, 3, 0, 1.0);

    // (u, v) = (0, 0)
    u = 0.0; v= 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 0, 0, u);
    cvmSet(src_points, 0, 1, v);
    cvmSet(dst_points, 0, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 0, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (0, height-1)
    u = 0.0; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 1, 0, u);
    cvmSet(src_points, 1, 1, v);
    cvmSet(dst_points, 1, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 1, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, height-1)
    u = m_uiWidth - 1; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 2, 0, u);
    cvmSet(src_points, 2, 1, v);
    cvmSet(dst_points, 2, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 2, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, 0)
    u = m_uiWidth - 1; v = 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 3, 0, u);
    cvmSet(src_points, 3, 1, v);
    cvmSet(dst_points, 3, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 3, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    cvFindHomography(src_points, dst_points, matH_F2T[i]);
    cvInvert(matH_F2T[i], matH_T2F[i], CV_LU);
  }

  cvReleaseMat(&src_points);
  cvReleaseMat(&dst_points);
  cvReleaseMat(&image);
  cvReleaseMat(&world);
  cvReleaseMat(&matIN_inv_from);
  cvReleaseMat(&mat_Rc2w_invIN_from);
}



bool CViewInterpolation::init_3Dwarp( double Z_near, double Z_far ) {
  double tableD2Z[256];
  double distance, temp1, temp2;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::init_3Dwarp()\n");
  #endif

  if ( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V, m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1= 1.0/Z_near - 1.0/Z_far;
  temp2= 1.0/Z_far;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  makeHomography( m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1], INIT );

  m_dBaselineDistance = pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance = sqrt(m_dBaselineDistance);
  return true;
}



bool CViewInterpolation::reset_3Dwarp(
                                    double near_C, double far_C, double trans_C,
                                    double near_V, double far_V, double trans_V
) {
  double tableD2Z[256];
  double distance, temp1, temp2;
  double fB;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::reset_3Dwarp()\n");
  #endif

  Mat_Trans_C[0]= trans_C;
  Mat_Trans_V[0]= trans_V;
  if( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V,
          m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1= 1.0/near_C - 1.0/far_C;
  temp2= 1.0/far_C;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  makeHomography( m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1] );

  m_dBaselineDistance=  pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance= sqrt(m_dBaselineDistance);

  fB = fabs(Mat_In_C[0]) * fabs(Mat_Trans_C[0] - Mat_Trans_V[0]);
  m_dMaxDisp = fB / fabs(near_C); // - fB / fabs(far_C);
  m_dispA = 1.0 / 255.0 * fB * fabs(1.0 / near_C - 1.0 / far_C);
  m_dispB = fB / fabs(far_C);

  return true;
}



unsigned char CViewInterpolation::weighted2( unsigned char src1, unsigned char src2, unsigned int steps, unsigned int count ) {
  return ( src1 * (steps-count) / steps ) + ( src2 * count / steps );
}



unsigned char CViewInterpolation::weighted3( unsigned char src1, unsigned char src2, unsigned char src3, unsigned int steps, unsigned int count ) {
  return (( src1 * (steps-count) / steps ) + ( src2 * count / steps )) / 2 + src3 / 2;
}



unsigned int CViewInterpolation::simple2( unsigned char src1, unsigned char src2 ) {
  unsigned int work= ( src1 + src2 ) / 2;
  return work;
}



unsigned int CViewInterpolation::simple3( unsigned char src1, unsigned char src2, unsigned char src3 ) {
  unsigned int work= ( src1 + src2*2 + src3 ) / 4;
  return work;
}



#define HOLEFLAG -1
#define NOHOLE -1
void CViewInterpolation::commonholeInpaint( IplImage* pView, IplImage* pHole ) {
  unsigned int pos, steps;
  unsigned char startVar[3];
  int startPos= NOHOLE;
  unsigned char endVar[3];
  int endPos= NOHOLE;
  bool startE= false;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::commonholeInpaint()\n");
  #endif

  IplImage* v1= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v2= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v3= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  cvSplit( pView, v1, v2, v3, NULL );

  pos= 0;
  for ( unsigned int h= 0; h < pView->height; h++ ) {
    for ( unsigned int w= 0; w < pView->width; w++, pos++ ) {
      if ( pHole->imageData[pos] == HOLEFLAG ) {
        if ( startPos == NOHOLE ) {
          startPos= pos;
          if ( w == 0 ) {
            startE= true;
          } else {
            startVar[0]= (unsigned char)v1->imageData[pos-1];
            startVar[1]= (unsigned char)v2->imageData[pos-1];
            startVar[2]= (unsigned char)v3->imageData[pos-1];
          }
        }
      }
      if ( pHole->imageData[pos] != HOLEFLAG || w == pView->width - 1 ) {
        if ( startPos != NOHOLE && endPos == NOHOLE ) {
          endPos= pos-1;
          if ( w == pView->width - 1 ) {
            if ( startE ) {
              printf("\nFULL WIDTH COMMON HOLE DETECTED!!!\n");
              startPos= endPos= NOHOLE;
              startE= false;
              continue; // pass!
            } else {
              endVar[0]= (unsigned char)simple2( 128, startVar[0] );
              endVar[1]= (unsigned char)simple2( 128, startVar[1] );
              endVar[2]= (unsigned char)simple2( 128, startVar[2] );
            }
          } else {
            endVar[0]= (unsigned char)v1->imageData[pos];
            endVar[1]= (unsigned char)v2->imageData[pos];
            endVar[2]= (unsigned char)v3->imageData[pos];
            if ( startE ) {
              startVar[0]= (unsigned char)simple2( 128, endVar[0] );
              startVar[1]= (unsigned char)simple2( 128, endVar[1] );
              startVar[2]= (unsigned char)simple2( 128, endVar[2] );
            }
          }
          steps= endPos - startPos + 2;
          for ( unsigned int p= startPos, count= 1; p <= endPos; p++, count++ ) {
            v1->imageData[p]= (unsigned char)weighted2( startVar[0], endVar[0], steps, count );
            v2->imageData[p]= (unsigned char)weighted2( startVar[1], endVar[1], steps, count );
            v3->imageData[p]= (unsigned char)weighted2( startVar[2], endVar[2], steps, count );
          }
          startPos= endPos= NOHOLE;
          startE= false;
        }
      }
    }
  }

  // 縦方向 Simple LPF
  for ( unsigned int w= 0; w < pView->width; w++ ) {
    pos= w;
    for ( unsigned int h= 0; h < pView->height; h++, pos+=pView->width ) {
      if ( pHole->imageData[pos] == HOLEFLAG ) {
        if ( h == 0 ) { // top side
            v1->imageData[pos]= (unsigned char)( (v1->imageData[pos] + v1->imageData[pos+pView->width]) / 2 );
            v2->imageData[pos]= (unsigned char)( (v2->imageData[pos] + v2->imageData[pos+pView->width]) / 2 );
            v3->imageData[pos]= (unsigned char)( (v3->imageData[pos] + v3->imageData[pos+pView->width]) / 2 );
        } else if ( h == pView->height - 1 ) { // bottom side
            v1->imageData[pos]= (unsigned char)( (v1->imageData[pos] + v1->imageData[pos-pView->width]) / 2 );
            v2->imageData[pos]= (unsigned char)( (v2->imageData[pos] + v2->imageData[pos-pView->width]) / 2 );
            v3->imageData[pos]= (unsigned char)( (v3->imageData[pos] + v3->imageData[pos-pView->width]) / 2 );
        } else {
            v1->imageData[pos]= (unsigned char)simple3( v1->imageData[pos+pView->width], v1->imageData[pos], v1->imageData[pos-pView->width] );
            v2->imageData[pos]= (unsigned char)simple3( v2->imageData[pos+pView->width], v2->imageData[pos], v2->imageData[pos-pView->width] );
            v3->imageData[pos]= (unsigned char)simple3( v3->imageData[pos+pView->width], v3->imageData[pos], v3->imageData[pos-pView->width] );
        }
      }
    }
  }

  cvMerge( v1, v2, v3, NULL, pView );

  SAFE_RELEASE_IMAGE( v1 );
  SAFE_RELEASE_IMAGE( v2 );
  SAFE_RELEASE_IMAGE( v3 );
}



void CViewInterpolation::depthInpaint( IplImage* map, IplImage* hole ) {
  int pos= 0;
  bool area= false;
  bool work= false;
  int start_w, stop_w, paint_pos; //, back_w, backpos;
  unsigned int left_value, right_value, paint_value;

  //DEBUG printf("\nheight:%i, width:%i\n", map->height, map->width);
  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::depthInpaint()\n");
  #endif

  for ( int h= 0; h < map->height; h++ ) {
    area= false;
    work= false;
    for ( int w= 0; w < map->width; w++, pos++ ) {
      if ( !area && hole->imageData[pos] == -1 ) { // 255
        area= true;
        start_w= w;
        if ( start_w == 0 )
          left_value= 255; // 255; -1
        else
          left_value= (unsigned char)map->imageData[pos-1];
      } else if ( area && hole->imageData[pos] == 0 ) {
        area= false;
        work= true;
        stop_w= w - 1;
        right_value= (unsigned char)map->imageData[pos];
      }
      if ( area && (w == (map->width - 1)) ) {
        work= true;
        area= false;
        stop_w= w;
        right_value= 255;
      }

      if ( work ) {
        if ( left_value <= right_value )
          paint_value= left_value;
        else
          paint_value= right_value;
//        printf("\n%i: h:%i, w:%i, start:%i, stop:%i, paint:%i, d:%i, h:%i", pos, h, w, start_w, stop_w, paint_value, map->imageData[pos], hole->imageData[pos]);
        for ( int paint_w= start_w; paint_w <= stop_w; paint_w++ ) {
          paint_pos= map->width * h + paint_w;
          //DEBUG printf(", %i", paint_pos);
          map->imageData[paint_pos]= paint_value;
          hole->imageData[paint_pos]= 0;
        }
        work= false;
        area= false;
      }
    }
  }
}



bool CViewInterpolation::setDepth(BYTE **pDepthMap) {
  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::setDepth()\n");
  #endif

  for(unsigned int h=0; h<m_uiHeight; h++)
    for(unsigned int w=0; w<m_uiWidth; w++)
      if (pDepthMap != NULL) m_pVirtualDepth[h][w]= pDepthMap[h][w];

  return true;
}



bool CViewInterpolation::setView(BYTE ***src) {
  unsigned int ptv;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::setView()\n");
  #endif

  for(unsigned int h=0; h<m_uiHeight; h++)
    for(unsigned int w=0; w<m_uiWidth; w++) {
      ptv = w + h * m_uiWidth;
      m_imgVirtualImage->imageData[ptv * 3     ] = src[0][h][w];
      m_imgVirtualImage->imageData[ptv * 3 + 1 ] = src[1][h][w];
      m_imgVirtualImage->imageData[ptv * 3 + 2 ] = src[2][h][w];
    }

  return true;
}



// Shift for Depth & Hole image.
bool CViewInterpolation::depthShift( BYTE **pDepthMap, unsigned int from ) {
  int h, w, v, u, PREu, ypos, start, end, step;
  bool isLEFT= false;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::depthShift()\n");
  #endif

  CvMat* m = cvCreateMat(3, 1, CV_64F);
  CvMat* mv = cvCreateMat(3, 1, CV_64F);

  cvZero(m_imgVirtualDepth);
  cvZero(m_imgSuccessSynthesis);

  isLEFT= ( from == LEFT );
  if ( isLEFT ) {
    start= 0;
    end= m_uiWidth;
    step= 1;
    PREu= -1;
  } else {
    start= m_uiWidth-1;
    end= 0;
    step= -1;
    PREu= m_uiWidth;
  }

  for(h=0; h<m_uiHeight; h++) {
    for ( w=start; ; w=w+step ) {
      if ( isLEFT && !(w < end) ) break;
      if ( !isLEFT && !(w >= end) ) break;

      cvmSet(m, 0, 0, w);
      cvmSet(m, 1, 0, h);
      cvmSet(m, 2, 0, 1);

      cvMatMulAdd( m_matH_R2V[ pDepthMap[h][w] ], m, 0, mv );
      u= mv->data.db[0] / mv->data.db[2] + 0.5;
      v= mv->data.db[1] / mv->data.db[2] + 0.5;

      if( u>=0 && u<m_uiWidth && v>=0 && v<m_uiHeight && m_pVirtualDepth[v][u] <= pDepthMap[h][w] ) { //
        m_pVirtualDepth[v][u] = pDepthMap[h][w];
        m_pSuccessSynthesis[v][u] = 255;
      }

      if ( ( isLEFT && PREu+1<u ) || ( !isLEFT && PREu-1>u ) ) {
        for ( ypos= PREu+step;; ypos= ypos+step ) {
          if ( isLEFT && !(ypos < u) ) break;
          if ( !isLEFT && !(ypos > u) ) break;
          if ( ypos < 0 ) break;
          if ( ypos >= m_uiWidth ) break;
          m_pVirtualDepth[v][ypos]= pDepthMap[h][w];
          m_pSuccessSynthesis[v][ypos]= 255;
        }
      }

      PREu= u;
    }
  }

  cvNot(m_imgSuccessSynthesis, m_imgHoles);

  cvReleaseMat(&m);
  cvReleaseMat(&mv);

  return true;
}



bool CViewInterpolation::viewsynthesis_reverse_3Dwarp(BYTE ***src, unsigned int from, IplImage* pHole)
{
  const int DEPTHGAP = 8;  //// 10 20NG 8 25:Glay出ないが、二重厳しい 16 8 Sep11 -> 4 -> 8 -> 6 // c*WeightRight
  const double DEPTHALLOW = 0.5;
  const int TH_U = 30;
  const int LIP_B = 3;
  const int HoleSearchStep = 16; // 16

  int ptv;
  unsigned int maxWidth = m_uiWidth * m_uiPrecision;

  int u, v, start, end, step; // , ref;
  bool isLEFT= false;
  bool isResidual= (pHole != NULL);
  bool validResidual= false;
  bool isPass;

  double c, d, disp, dMaxDisp;
  int nearU1, nearU2;
  int u1, ud;
  double uf, w0, w1;
  int diffU, preU, preU3m, preU3p, preW, delta;

  int iOver;
  bool inHole;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::viewsynthesis_reverse_3Dwarp()\n");
  #endif

  CvMat* m = cvCreateMat(3, 1, CV_64F);
  CvMat* mv = cvCreateMat(3, 1, CV_64F);

  //if ( !isResidual )
  cvSet( m_imgVirtualImage, CV_RGB(128, 128, 128) ); // 0, 128, 128
  cvSet( m, cvScalarAll(0) );
  cvSet( mv, cvScalarAll(0) );
  cvSet( m_imgSuccessSynthesis, cvScalarAll(255) );

  isLEFT= ( from == LEFT );
  if ( isResidual ) { isLEFT= !isLEFT; }
  if ( isLEFT ) {
    start= m_uiWidth-1;
    end= 0;
    step= -1;
  } else {
    start= 0;
    end= m_uiWidth-1;
    step= 1;
  }
  // if ( isResidual ) ref= -step; else ref= step;

  for(int h=0; h<m_uiHeight; h++) {
    iOver= 0;
    inHole= false;
    preU= start;
    validResidual= false;

    for(int w=start; ; w=w+step) {

      if ( isLEFT && (w < end) ) break;
      if ( !isLEFT && (w > end) ) break;
      if ( isResidual && pHole->imageData[h*m_uiWidth + w] == 0 ) {
        validResidual= false;
        inHole= false;
        continue;
      }
      isPass= false;

      dMaxDisp= m_dMaxDisp;

      do {
        if ( isLEFT ) {
          nearU1= w - dMaxDisp;
          if (nearU1 >= 0) {
            c= (m_pVirtualDepth[h][nearU1] - m_pVirtualDepth[h][w]); // 近傍画素位置１と現在位置のデプス値差分
            disp= m_dispA * c; //  + m_dispB;
            if (c > DEPTHGAP) {
              if ( disp > dMaxDisp ) { isPass= true; break; } /////////////////////////////////
              nearU2= w - disp; // * m_weightVR; // 近傍画素位置二 ＝ 現在位置＋デプス差分;
              if ( nearU2 >= 0 ) {
                d= (m_pVirtualDepth[h][nearU2] - m_pVirtualDepth[h][w]); // 近傍画素位置２と現在位置のデプス値差分
                c*= DEPTHALLOW; // 近傍画素位置１と現在位置のデプス値差分をマイルドに
                if (d >= c) { isPass= true; break; } /////////////////////////////////
              }
            }
          }
        } else {
          nearU1= w + dMaxDisp;
          if (nearU1<(int)m_uiWidth) {
            c= (m_pVirtualDepth[h][nearU1] - m_pVirtualDepth[h][w]); // 近傍画素位置１と現在位置のデプス値差分
            disp= m_dispA * c; //  + m_dispB;
            if (c > DEPTHGAP) {
              if ( disp > dMaxDisp ) { isPass= true; break; } /////////////////////////////////
              nearU2= w + disp; // * m_weightVR; // 近傍画素位置二 ＝ 現在位置＋デプス差分
              if (nearU2 < (int)m_uiWidth) {
                d= (m_pVirtualDepth[h][nearU2] - m_pVirtualDepth[h][w]); // 近傍画素位置２と現在位置のデプス値差分
                c*= DEPTHALLOW; // 近傍画素位置１と現在位置のデプス値差分をマイルドに
                if (d >= c) { isPass= true; break; } /////////////////////////////////
              }
            }
          }
        }
        dMaxDisp-= m_dMaxDisp / HoleSearchStep;
      } while( dMaxDisp > 0 );


      ptv = w + h * m_uiWidth;
      cvmSet(mv, 0, 0, w);
      cvmSet(mv, 1, 0, h);
      cvmSet(mv, 2, 0, 1);
      cvMatMulAdd( m_matH_V2R[m_pVirtualDepth[h][w]], mv, 0, m );
      // Jun1,2012 OUT   u = m->data.db[0] * m_uiPrecision / m->data.db[2] + 0.5;
      uf= m->data.db[0] * m_uiPrecision / m->data.db[2];
      u= (int)uf;
      u1= u + 1; if ( u1 >= maxWidth ) u1= maxWidth - 1;
      w1= uf - u;
      w0= 1.0 - w1;
      v = m->data.db[1] / m->data.db[2] + 0.5;

      if ( u>=0 && u<maxWidth && v>=0 && v<m_uiHeight ) {

        if (isPass) {
          m_pSuccessSynthesis[h][w]= 0;
        } else { // 穴でない
          inHole= false;
          validResidual= true;

          m_imgVirtualImage->imageData[ptv * 3     ] = w0 * src[0][v][u] + w1 * src[0][v][u1] + 0.5;
          m_imgVirtualImage->imageData[ptv * 3 + 1 ] = w0 * src[1][v][u] + w1 * src[1][v][u1] + 0.5;
          m_imgVirtualImage->imageData[ptv * 3 + 2 ] = w0 * src[2][v][u] + w1 * src[2][v][u1] + 0.5;
        }

        preU= u;
      } else {
        m_pSuccessSynthesis[h][w]=0;
        iOver= 1;
      }

    }
  }

  cvReleaseMat(&m);
  cvReleaseMat(&mv);

  cvNot(m_imgSuccessSynthesis, m_imgHoles); // pixels which couldn't be synthesized

  return true;
}



int CViewInterpolation::DoOneFrame (
          CIYuv *pSynYuvBuffer, double near_T1, double far_T1, double trans_T1
) {
  unsigned int h, pos;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::DoOneFrame()\n");
  #endif

  IplImage* imgWork= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  IplImage* imgResult2= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  IplImage* imgResult= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  IplImage* imgMask= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* imgDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* imgCommonHole= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* imgExpandHole= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* imgXorHole= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  // DEBUG Jpeg OUT
  char fname[40];
  int params[] = {CV_IMWRITE_JPEG_QUALITY, 70, -1};
  IplImage* imgRGB= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  /**/

  if (VariParam) {
    if( !m_pS->reset_3Dwarp( near_V1, far_V1, trans_V1, near_T1, far_T1, trans_T1 ) ) return false;
    if( !m_pV->reset_3Dwarp( near_C2, far_C2, trans_C2, near_T1, far_T1, trans_T1 ) ) return false;
    if( !m_pR->reset_3Dwarp( near_C1, far_C1, trans_C1, near_T1, far_T1, trans_T1 ) ) return false;
    printf(" V ");
  }

  // Step.1  Depth ///////////////////////////////////////////////
  m_pD->setDepth( Depth );
  cvSmooth( m_pD->getVirtualDepthMap(), m_pD->getWorkDepthMap(), CV_MEDIAN, 5 );


  if ( !m_pS->depthShift( m_pD->m_pWorkDepth, from ) ) return false;

  cvSmooth( m_pS->getVirtualDepthMap(), imgDepth, CV_MEDIAN, 3 );
  cvSmooth( m_pS->getHolePixels(), imgMask, CV_MEDIAN, 3 );
  depthInpaint( imgDepth, imgMask );
  if ( debug ) {
    sprintf( fname, "%03iDsmooth%1i_%s.png", fnum, from, m_pV->getVirtualCameraName().c_str() );
    cvSaveImage( fname, imgDepth );
    // sprintf( fname, "%03iD%1i_%s.png", fnum, from, m_pV->getVirtualCameraName().c_str() );
    // cvSaveImage( fname, m_pS->getVirtualDepthMap() );
  }

  for( h=pos=0; h<m_pV->m_uiHeight; h++, pos+=imgDepth->widthStep ) {
    m_pV->m_pVirtualDepth[h] = (BYTE *) &(imgDepth->imageData[pos]);
    m_pR->m_pVirtualDepth[h] = (BYTE *) &(imgDepth->imageData[pos]);
  }

  // Step.2  View ///////////////////////////////////////////////
  if ( !m_pV->viewsynthesis_reverse_3Dwarp(ViewM, from) ) return false;
  cvZero(imgResult);
  cvCopy( m_pV->getVirtualImage(), imgResult ); // NoMask Jun12,'13 , m_pV->getSynthesizedPixels() );
  if ( debug ) {
    // drawHoledge( imgResult, m_pV->getSynthesizedPixels(), -1, 0, 0 ); // for DEBUG: View GREEN
    // drawHoledge( imgResult, m_pV->getHolePixels(), -1, -1, -1 ); // for DEBUG: Hole PINK

    cvCvtColor( imgResult, imgRGB, CV_YCrCb2RGB );
    sprintf( fname, "%03iV%1i_%s.jpg", fnum, from, m_pV->getVirtualCameraName().c_str() );
    cvSaveImage( fname, imgRGB, params );
  }

  // Step.3  Residual ///////////////////////////////////////////////
  //m_pR->setView( ViewR );
  cvDilate( m_pV->getHolePixels(), imgExpandHole, NULL, 1 );
  // cvSmooth( m_pV->getHolePixels(), imgExpandHole, CV_MEDIAN, 3 );
  // cvXor( m_pV->getHolePixels(), imgExpandHole, imgXorHole );
  // cvSmooth( imgResult, imgWork, CV_MEDIAN, 3 );
  // cvCopy( imgWork, imgResult, imgXorHole );

  if ( !m_pR->viewsynthesis_reverse_3Dwarp(ViewR, from, imgExpandHole) ) return false;
  cvCopy( m_pR->getVirtualImage(), imgResult, imgExpandHole );
  if ( debug ) {
    cvCvtColor( m_pR->getVirtualImage(), imgRGB, CV_YCrCb2RGB );
    sprintf( fname, "%03iR%1i_%s.jpg", fnum, from, m_pV->getVirtualCameraName().c_str() );
    cvSaveImage( fname, imgRGB, params );
  }

  // Step.4  Common Hole ///////////////////////////////////////////////
//  cvAnd( m_pR->getHolePixels(), m_pV->getHolePixels(), imgCommonHole ); // imgExpandHole, imgCommonHole );
  cvInpaint( imgResult, m_pR->getHolePixels(), imgResult2, 1, CV_INPAINT_NS );

  looseHoledge( imgResult2, imgExpandHole );
  looseHoledge( imgResult2, m_pR->getHolePixels() ); // imgCommonHole );

  if ( debug ) {
    // drawHoledge( imgResult2, imgExpandHole, -1, 0, 0 ); // for DEBUG: Residuals GREEN
    // drawHoledge( imgResult2, m_pR->getHolePixels(), -1, -1, -1 ); // for DEBUG: Common Hole PINK

    cvCvtColor( imgResult2, imgRGB, CV_YCrCb2RGB );
    sprintf( fname, "%03iVR%1i_%s.jpg", fnum, from, m_pV->getVirtualCameraName().c_str() );
    cvSaveImage( fname, imgRGB, params );
  }

  pSynYuvBuffer->setDataFromImgYUV( imgResult2 );

  cvReleaseImage( &imgWork );
  cvReleaseImage( &imgResult );
  cvReleaseImage( &imgResult2 );
  cvReleaseImage( &imgMask );
  cvReleaseImage( &imgDepth );
  cvReleaseImage( &imgCommonHole );
  cvReleaseImage( &imgExpandHole );
  cvReleaseImage( &imgXorHole );
  cvReleaseImage( &imgRGB ); // DEBUG

  return true;
}



#define TH 20
void CViewInterpolation::looseHoledge( IplImage* pView, IplImage* pHole ) {
  unsigned int pos;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::looseHoledge()\n");
  #endif

  IplImage* v1= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v2= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v3= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  cvSplit( pView, v1, v2, v3, NULL );

  pos= pView->width;
  for ( unsigned int h= 1; h < pView->height-2; h++ ) {
    for ( unsigned int w= 0; w < pView->width; w++, pos++ ) {
      if ( pHole->imageData[pos] != pHole->imageData[pos + pView->width] ) {
        if ( abs((unsigned char)v1->imageData[pos] - (unsigned char)v1->imageData[pos + pView->width]) < TH ) {
          v1->imageData[pos]= ( (unsigned char)v1->imageData[pos - pView->width] + (unsigned char)v1->imageData[pos] * 2 + (unsigned char)v1->imageData[pos + pView->width] + 2 ) >> 2;
          v1->imageData[pos + pView->width]= ( (unsigned char)v1->imageData[pos] + (unsigned char)v1->imageData[pos + pView->width] * 2 + (unsigned char)v1->imageData[pos + pView->width*2] + 2 ) >> 2;
        }
      }
    }
  }

  pos= 0;
  for ( unsigned int h= 0; h < pView->height; h++ ) {
    pos++;
    for ( unsigned int w= 1; w < pView->width-2; w++, pos++ ) {
      if ( pHole->imageData[pos] != pHole->imageData[pos + 1] ) {
        if ( abs((unsigned char)v1->imageData[pos] - (unsigned char)v1->imageData[pos + 1]) < TH ) {
          v1->imageData[pos]= ( (unsigned char)v1->imageData[pos - 1] + (unsigned char)v1->imageData[pos] * 2 + (unsigned char)v1->imageData[pos + 1] + 2 ) >> 2;
          v1->imageData[pos + 1]= ( (unsigned char)v1->imageData[pos] + (unsigned char)v1->imageData[pos + 1] * 2 + (unsigned char)v1->imageData[pos + 2] + 2 ) >> 2;
        }
      }
    }
    pos+= 2;
  }

  cvMerge( v1, v2, v3, NULL, pView );

  cvReleaseImage( &v1 );
  cvReleaseImage( &v2 );
  cvReleaseImage( &v3 );
}



// for DEBUG USE
void CViewInterpolation::drawHoledge( IplImage* pView, IplImage* pHole, char Y, char U, char V ) {
  unsigned int pos;

  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::drawHoledge()\n");
  #endif

  IplImage* v1= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v2= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  IplImage* v3= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  cvSplit( pView, v1, v2, v3, NULL );

  pos= pView->width;
  for ( unsigned int h= 1; h < pView->height-2; h++ ) {
    for ( unsigned int w= 0; w < pView->width; w++, pos++ ) {
      if ( pHole->imageData[pos] != pHole->imageData[pos + pView->width] ) {
        v1->imageData[pos]= Y;
        v2->imageData[pos]= U;
        v3->imageData[pos]= V;
      }
    }
  }

  pos= 0;
  for ( unsigned int h= 0; h < pView->height; h++ ) {
    pos++;
    for ( unsigned int w= 1; w < pView->width-2; w++, pos++ ) {
      if ( pHole->imageData[pos] != pHole->imageData[pos + 1] ) {
        v1->imageData[pos]= Y;
        v2->imageData[pos]= U;
        v3->imageData[pos]= V;
      }
    }
    pos+= 2;
  }

  cvMerge( v1, v2, v3, NULL, pView );

  cvReleaseImage( &v1 );
  cvReleaseImage( &v2 );
  cvReleaseImage( &v3 );
}



void CViewInterpolation::ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize)
{
  #ifdef PTRACE
    printf("TRACE: CViewInterpolation::ToFilterDepth()\n");
  #endif

  IplImage* depthSuccess= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1); // Jul4,2011 NiCT;
  IplImage* depthMask= cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);;

  // デプス穴マップSRCからデプス成功マップを作成
  // デプス穴マップSRCにMFかけてデプス穴マップDSTを作成
  // デプス成功マップとデプス穴マップDSTのANDをとってデプス有マスク作成
  // デプス有マスクをかけて、デプス穴マップSRCからデプス穴マップDSTにコピー
  cvNot(pDepthNEGSRC, depthSuccess);
  cvSmooth(pDepthNEGSRC, pDepthNEGDST, CV_MEDIAN, fsize);
  cvAnd(depthSuccess, pDepthNEGDST, depthMask);
  cvCopy(pDepthNEGSRC, pDepthNEGDST, depthMask);

  // デプスSRCに」MDかけてデプスDSTを作成
  // デプス有マスクをかけて、デプスSRCからデプスDSTにコピー
  cvSmooth(pDepthSRC, pDepthDST, CV_MEDIAN, fsize);
  cvCopy(pDepthSRC, pDepthDST, depthMask);

  cvReleaseImage( &depthSuccess );
  cvReleaseImage( &depthMask );
}
