// View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 7 64bit with Visual Studio 2010 C++ x64
//    OpenCV 2.4.4 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: Apr, 2013.

#include "DepthSynthesis.h"
#include "SubMixOptions.h"

// for OpenCV 2.1.0 include <opencv/cvaux.h>
#include <opencv2/legacy/compat.hpp> // for cvmMul()
#include <opencv2/calib3d/calib3d.hpp> // for cvFindHomography()
#include <opencv2/imgproc/imgproc_c.h> // for cvSmooth(), cvDilate()
// #include <opencv2/photo/photo_c.h> // for cvInpaint()
#include <opencv2/highgui/highgui_c.h> // for DEBUG cvSaveImage()


#ifndef SAFE_RELEASE_IMAGE
#define SAFE_RELEASE_IMAGE(p) { if((p)!=NULL){ cvReleaseImage(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_RELEASE_MAT
#define SAFE_RELEASE_MAT(p) { if((p)!=NULL){ cvReleaseMat(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_FREE
#define SAFE_FREE(p) { if((p)!=NULL){ free(p); (p)=NULL; } }
#endif


CDepthInterpolationGeneral::CDepthInterpolationGeneral() {
  int i;

  m_uiPrecision = 1;

  m_imgVirtualDepth = NULL;
  m_imgSuccessSynthesis = m_imgHoles = m_imgBound = m_imgJump = NULL;
  m_pVirtualDepth = m_pSuccessSynthesis = NULL;
  m_pcDepthSynthesisLeft = NULL;
  m_pcDepthSynthesisRight = NULL;
  m_pcDepthSynthesisOutside = NULL;
  m_imgJump= NULL;
  m_pJump = NULL;

  m_depthBlended0= m_depthBlended1= m_depthBlended2= NULL;
  m_depthPre1= m_depthPre2= NULL;
  m_depthSuccess= NULL;

  m_imgLHole= m_imgLSuccess= m_imgRHole= m_imgRSuccess= NULL;
  m_imgOHole= m_imgOSuccess= NULL;
  m_imgComHole= NULL;

  for(i=0; i<5; i++)
    m_imgTemp[i] = m_imgMask[i] = NULL;

  for(i=0; i<256; i++)
    m_matH_R2V[i] = m_matH_V2R[i] = NULL;

  // for VariParam
  for(i=0; i<2; i++)
    m_Mat_in[i]= m_Mat_ex_c2w[i]= m_Mat_proj_w2i[i]= NULL;

  cv::setUseOptimized(true);
  //if (cv::useOptimized()) printf("\nuseOptimized(): true"); else printf("\nuseOptimized(): false");
}



CDepthInterpolationGeneral::~CDepthInterpolationGeneral()
{
  xReleaseMemory();
}



void CDepthInterpolationGeneral::xReleaseMemory () {
  int i;

  SAFE_RELEASE_IMAGE(m_imgVirtualDepth);
  SAFE_RELEASE_IMAGE(m_imgSuccessSynthesis);
  SAFE_RELEASE_IMAGE(m_imgJump);
  SAFE_RELEASE_IMAGE(m_imgHoles);
  SAFE_RELEASE_IMAGE(m_imgBound);
  SAFE_RELEASE_IMAGE(m_depthBlended0);
  SAFE_RELEASE_IMAGE(m_depthBlended1);
  SAFE_RELEASE_IMAGE(m_depthBlended2);
  SAFE_RELEASE_IMAGE(m_depthBlendedMix);
  SAFE_RELEASE_IMAGE(m_depthBlendedMix2);
  SAFE_RELEASE_IMAGE(m_depthSuccess);

  SAFE_RELEASE_IMAGE(m_imgLHole);
  SAFE_RELEASE_IMAGE(m_imgLSuccess);
  SAFE_RELEASE_IMAGE(m_imgRHole);
  SAFE_RELEASE_IMAGE(m_imgRSuccess);
  SAFE_RELEASE_IMAGE(m_imgOHole);
  SAFE_RELEASE_IMAGE(m_imgOSuccess);
  SAFE_RELEASE_IMAGE(m_imgComHole);

  if(m_pcDepthSynthesisLeft!=NULL) delete m_pcDepthSynthesisLeft;
  if(m_pcDepthSynthesisRight!=NULL) delete m_pcDepthSynthesisRight;
  if(m_pcDepthSynthesisOutside!=NULL) delete m_pcDepthSynthesisOutside;
  m_pcDepthSynthesisLeft = m_pcDepthSynthesisRight = m_pcDepthSynthesisOutside = NULL;

  for (i=0; i<5; i++) {
    SAFE_RELEASE_IMAGE(m_imgTemp[i]);
    SAFE_RELEASE_IMAGE(m_imgMask[i]);
  }

  SAFE_FREE(m_pVirtualDepth);
  SAFE_FREE(m_pSuccessSynthesis);
  SAFE_FREE(m_pJump);

  for(i=0; i<256; i++) {
    SAFE_RELEASE_MAT( m_matH_R2V[i] );
    SAFE_RELEASE_MAT( m_matH_V2R[i] );
  }

  // for VariParam
  for(i=0; i<2; i++) {
    SAFE_RELEASE_MAT( m_Mat_in[i] );
    SAFE_RELEASE_MAT( m_Mat_ex_c2w[i] );
    SAFE_RELEASE_MAT( m_Mat_proj_w2i[i] );
  }
}



void CDepthInterpolationGeneral::convertCameraParam(CvMat *exMat_dst, CvMat *exMat_src)
{
  int i, j;
  double val;

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      cvmSet(exMat_dst, i, j, cvmGet(exMat_src, j, i));

  for(i=0; i<3; i++) {
    val = 0.0;
    for(j=0; j<3; j++)
      val -= cvmGet(exMat_dst, i, j)*cvmGet(exMat_src, j, 3);
    cvmSet(exMat_dst, i, 3, val);
  }
}



// Init both(all) depth channel.
bool CDepthInterpolationGeneral::InitLR(
        unsigned int uiWidth, unsigned int uiHeight, bool variparam, unsigned int mode,
        double dZnearL, double dZfarL, double dZnearR, double dZfarR, double dZnearO, double dZfarO,
        const char *strRefLCamID, const char *strRefRCamID, const char *strRefOCamID, const char *strVirCamID,
        double Mat_In_Left[9], double Mat_Ex_Left[9], double Mat_Trans_Left[3],
        double Mat_In_Right[9], double Mat_Ex_Right[9], double Mat_Trans_Right[3],
        double Mat_In_Outside[9], double Mat_Ex_Outside[9], double Mat_Trans_Outside[3],
        double Mat_In_Virtual[9], double Mat_Ex_Virtual[9], double Mat_Trans_Virtual[3]
) {
  VariParam= variparam;
  #ifdef DEBUGWRITE
  printf( "DEBUG: InitLR() Start.\n" );
  #endif

  m_pcDepthSynthesisLeft = new CDepthInterpolationGeneral();
  m_pcDepthSynthesisRight = new CDepthInterpolationGeneral();
  if(m_pcDepthSynthesisLeft==NULL || m_pcDepthSynthesisRight==NULL) return false;

  // The number of arguments of Init increased.
  // Intrinsic and Extrinsic parameters of (left or right) and virtual cameras are set throught this function.
  if(!m_pcDepthSynthesisLeft->Init(uiWidth, uiHeight, variparam,
                                    dZnearL, dZfarL,
                                    strRefLCamID, strVirCamID,
                                    Mat_In_Left, Mat_Ex_Left, Mat_Trans_Left,
                                    Mat_In_Virtual, Mat_Ex_Virtual, Mat_Trans_Virtual
                                    ))  return false;
  #ifdef DEBUGWRITE
  printf( "DEBUG: m_pcDepthSynthesisLeft->Init() done.\n" );
  #endif

  if(!m_pcDepthSynthesisRight->Init(uiWidth, uiHeight, variparam,
                                    dZnearR, dZfarR,
                                    strRefRCamID, strVirCamID,
                                    Mat_In_Right, Mat_Ex_Right, Mat_Trans_Right,
                                    Mat_In_Virtual, Mat_Ex_Virtual, Mat_Trans_Virtual
                                    ))  return false;
  #ifdef DEBUGWRITE
  printf( "DEBUG: m_pcDepthSynthesisRight->Init() done.\n" );
  #endif

  if ( mode != QUARTER && mode != HALF_QUARTER ) {
    m_pcDepthSynthesisOutside = new CDepthInterpolationGeneral();
    if(m_pcDepthSynthesisOutside==NULL) return false;
    if(!m_pcDepthSynthesisOutside->Init(uiWidth, uiHeight, variparam,
                                    dZnearO, dZfarO,
                                    strRefOCamID, strVirCamID,
                                    Mat_In_Outside, Mat_Ex_Outside, Mat_Trans_Outside,
                                    Mat_In_Virtual, Mat_Ex_Virtual, Mat_Trans_Virtual
                                    ))  return false;
    #ifdef DEBUGWRITE
    printf( "DEBUG: m_pcDepthSynthesisOutside->Init() done.\n" );
    #endif
  } else {
    // 2View: QUARTER, HALF_QUARTER: NOP
    /////// ??  Global Depth: QUARTER_ODD, QUARTER_ODD_L: NOP
  }

  m_depthBlended0  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_depthBlended1  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_depthBlended2  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_depthBlendedMix  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_depthBlendedMix2  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgMask[2]  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgMask[3]  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgMask[4]  = cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_depthSuccess= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);

  m_imgLHole= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgLSuccess= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgRHole= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgRSuccess= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgOHole= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgOSuccess= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);
  m_imgComHole= cvCreateImage(cvSize(uiWidth, uiHeight), 8, 1);

  #ifdef DEBUGWRITE
  printf( "DEBUG: InitLR() done.\n" );
  #endif
  return true;
}



bool CDepthInterpolationGeneral::Init(
        unsigned int uiWidth, unsigned int uiHeight, bool variparam,
        double dZnear, double dZfar,
        const char *strRefCamID, const char *strVirCamID,
        double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
        double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
) {
  unsigned int h, pos;

  //xReleaseMemory();

  VariParam= variparam;
  m_uiHeight = uiHeight;
  m_uiWidth = uiWidth;
  m_uiPicsize = m_uiHeight*m_uiWidth;

  m_imgVirtualDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgSuccessSynthesis = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgJump = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  m_imgHoles    = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgBound    = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgMask[0]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgMask[1]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgTemp[0]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgTemp[1]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  if((m_pVirtualDepth = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgVirtualDepth->widthStep)
    m_pVirtualDepth[h] = (BYTE *) &(m_imgVirtualDepth->imageData[pos]);

  if((m_pSuccessSynthesis = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgSuccessSynthesis->widthStep)
    m_pSuccessSynthesis[h] = (BYTE *) &(m_imgSuccessSynthesis->imageData[pos]);

  if((m_pJump = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgJump->widthStep)
    m_pJump[h] = (BYTE *) &(m_imgJump->imageData[pos]);

//  if (VariParam) {
  for(unsigned int i=0; i<2; i++) {
    m_Mat_in[i] = cvCreateMat(3, 3, CV_64F); //intrinsic parameter of camera (3x3 matrix)
    m_Mat_ex_c2w[i] = cvCreateMat(3, 4, CV_64F); //extrinsic parameter of camera (3x4 matrix)
    m_Mat_proj_w2i[i] = cvCreateMat(3, 4, CV_64F); // projection matrix
  }

  for (unsigned int i=0; i<9; i++) Mat_In_C[i]= Mat_In_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_C[i]= Mat_Ex_Ref[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_C[i]= Mat_Trans_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_In_V[i]= Mat_In_Vir[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_V[i]= Mat_Ex_Vir[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_V[i]= Mat_Trans_Vir[i];
//  }

  #ifdef DEBUGWRITE
  printf( "DEBUG: CDepthInterpolationGeneral::Init() done, without init_3Dwarp().\n" );
  #endif
  return init_3Dwarp ( dZnear, dZfar );
}



bool CDepthInterpolationGeneral::init_camera_param (
                                    double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
                                    double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3],
                                    CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]
) {
  int i, j;
  CvMat *mat_ex_w2c;

  for(i=0; i<3; i++)
    for(j=0; j<3;j++)
    {
      mat_in[0]->data.db[j+3*i] = *(Mat_In_Ref+3*i+j);
      mat_in[1]->data.db[j+3*i] = *(Mat_In_Vir+3*i+j);
    }

  for(i=0; i<3; i++)
    for(j=0; j<3;j++)
    {
      mat_ex_c2w[0]->data.db[j+4*i] = *(Mat_Ex_Ref+3*i+j);
      mat_ex_c2w[1]->data.db[j+4*i] = *(Mat_Ex_Vir+3*i+j);
    }

  for(i=0; i<3; i++)
  {
    mat_ex_c2w[0]->data.db[3+4*i] = *(Mat_Trans_Ref+i);
    mat_ex_c2w[1]->data.db[3+4*i] = *(Mat_Trans_Vir+i);
  }

  if(mat_proj!=NULL)
  {
    mat_ex_w2c = cvCreateMat(3, 4, CV_64F);
    for(i=0; i<2; i++)
    {
      convertCameraParam(mat_ex_w2c, mat_ex_c2w[i]);
      cvmMul(mat_in[i], mat_ex_w2c, mat_proj[i]);  // Proj = inMat_c2i * exMat_w2c
    }
    cvReleaseMat(&mat_ex_w2c);
  }

  return true;
}



void CDepthInterpolationGeneral::image2world_with_z(
                                      CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from, CvMat *image, CvMat *world
) {
  CvMat *temp;
  double s;

  temp = cvCreateMat(3, 1, CV_64F);

  cvmMul(mat_Rc2w_invIN_from, image, temp);

  s = ( cvmGet(world, 2, 0) - cvmGet(matEX_c2w_from, 2, 3) ) / cvmGet(temp, 2, 0);

  cvmSet(world, 0, 0, s*cvmGet(temp, 0, 0) + cvmGet(matEX_c2w_from, 0, 3));
  cvmSet(world, 1, 0, s*cvmGet(temp, 1, 0) + cvmGet(matEX_c2w_from, 1, 3));

  cvReleaseMat(&temp);
}



void CDepthInterpolationGeneral::makeHomography(
                                    CvMat *matH_F2T[256], CvMat *matH_T2F[256], double adTable[256],
                                    CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to
) {
  int i, j, k;
  double val, u, v;
  CvMat *matIN_inv_from, *mat_Rc2w_invIN_from;

  CvMat *src_points  = cvCreateMat(4, 2, CV_64F);
  CvMat *dst_points  = cvCreateMat(4, 2, CV_64F);
  CvMat *image      = cvCreateMat(3, 1, CV_64F);
  CvMat *world      = cvCreateMat(4, 1, CV_64F);

  matIN_inv_from      = cvCreateMat(3, 3, CV_64F);
  mat_Rc2w_invIN_from  = cvCreateMat(3, 3, CV_64F);

  cvInvert(matIN_from, matIN_inv_from, CV_LU);
  for(i=0; i<3; i++)
  {
    for(j=0; j<3; j++)
    {
      val=0.0;
      for(k=0; k<3; k++)
      {
        val += cvmGet(matEX_c2w_from, i, k)*cvmGet(matIN_inv_from, k, j);
      }
      cvmSet(mat_Rc2w_invIN_from, i, j, val);
    }
  }

  for(i=0; i<256; i++)
  {
    // matH_F2T[i] = cvCreateMat(3, 3, CV_64F);
    // matH_T2F[i] = cvCreateMat(3, 3, CV_64F);

    cvmSet(world, 2, 0, adTable[i]);
    cvmSet(world, 3, 0, 1.0);

    // (u, v) = (0, 0)
    u = 0.0; v= 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 0, 0, u);
    cvmSet(src_points, 0, 1, v);
    cvmSet(dst_points, 0, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 0, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (0, height-1)
    u = 0.0; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 1, 0, u);
    cvmSet(src_points, 1, 1, v);
    cvmSet(dst_points, 1, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 1, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, height-1)
    u = m_uiWidth - 1; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 2, 0, u);
    cvmSet(src_points, 2, 1, v);
    cvmSet(dst_points, 2, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 2, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, 0)
    u = m_uiWidth - 1; v = 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 3, 0, u);
    cvmSet(src_points, 3, 1, v);
    cvmSet(dst_points, 3, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 3, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    cvFindHomography(src_points, dst_points, matH_F2T[i]);
    cvInvert(matH_F2T[i], matH_T2F[i], CV_LU);
  }

  cvReleaseMat(&src_points);
  cvReleaseMat(&dst_points);
  cvReleaseMat(&image);
  cvReleaseMat(&world);
  cvReleaseMat(&matIN_inv_from);
  cvReleaseMat(&mat_Rc2w_invIN_from);
}



bool CDepthInterpolationGeneral::init_3Dwarp( double Z_near, double Z_far ) {
  double tableD2Z[256];
  double distance, temp1, temp2;

  if( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V, m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1 = 1.0/Z_near - 1.0/Z_far;
  temp2 = 1.0/Z_far;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  for (unsigned int i=0; i<256; i++) {
    m_matH_R2V[i]= cvCreateMat(3, 3, CV_64F);
    m_matH_V2R[i]= cvCreateMat(3, 3, CV_64F);
  }

  makeHomography(m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1]);

  m_dBaselineDistance=  pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance = sqrt(m_dBaselineDistance);

  #ifdef DEBUGWRITE
  printf( "DEBUG: CDepthInterpolationGeneral::init_3Dwarp() done.\n" );
  #endif
  return true;
}



bool CDepthInterpolationGeneral::reset_3Dwarp(
                                    double near_C, double far_C, double trans_C,
                                    double near_V, double far_V, double trans_V
) {
  double tableD2Z[256];
  double distance, temp1, temp2;

  Mat_Trans_C[0]= trans_C;
  Mat_Trans_V[0]= trans_V;
  if( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V,
          m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1= 1.0 / near_C - 1.0 / far_C;
  temp2= 1.0 / far_C;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  makeHomography(m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1]);

  m_dBaselineDistance=  pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance= sqrt(m_dBaselineDistance);

  return true;
}



void CDepthInterpolationGeneral::depthInpaint( IplImage* map, IplImage* hole ) {
  int pos= 0;
  bool area= false;
  bool work= false;
  int start_w, stop_w, paint_pos; //, back_w, backpos;
  unsigned int left_value, right_value, paint_value;

  for ( int h= 0; h < map->height; h++ ) {
    area= false;
    work= false;
    for ( int w= 0; w < map->width; w++, pos++ ) {
      if ( !area && hole->imageData[pos] == -1 ) { // 255
        area= true;
        start_w= w;
        if ( start_w == 0 )
          left_value= 255; // 255; -1
        else
          left_value= (unsigned char)map->imageData[pos-1];
      } else if ( area && hole->imageData[pos] == 0 ) {
        area= false;
        work= true;
        stop_w= w - 1;
        right_value= (unsigned char)map->imageData[pos];
      }
      if ( area && (w == (map->width - 1)) ) {
        work= true;
        area= false;
        stop_w= w;
        right_value= 255;
      }

      if ( work ) {
        if ( left_value <= right_value )
          paint_value= left_value;
        else
          paint_value= right_value;

        for ( int paint_w= start_w; paint_w <= stop_w; paint_w++ ) {
          paint_pos= map->width * h + paint_w;
          map->imageData[paint_pos]= paint_value;
          hole->imageData[paint_pos]= 0;
        }
        work= false;
        area= false;
      }
    }
  }
}



void CDepthInterpolationGeneral::depthEnhanced( IplImage* map, IplImage* out, unsigned int th ) {
  unsigned int pos= 0;

  for ( int h= 0; h < map->height; h++ )
    for ( int w= 0; w < map->width; w++, pos++ )
      out->imageData[pos]= map->imageData[pos];

  pos= 0;
  for ( int h= 0; h < map->height; h++ )
    for ( int w= 0; w < map->width; w++, pos++ ) {
      if ( w+7 >= map->width ) {
        pos+= 7;
        break;
      }

      if (
        abs(map->imageData[pos] - map->imageData[pos+2]) < th
        && abs(map->imageData[pos+5] - map->imageData[pos+7]) < th
        && abs(map->imageData[pos+1] - map->imageData[pos+6]) > th // Sep4,'13 5 -> 6
      ) {
        if ( abs(map->imageData[pos+2] - map->imageData[pos+3]) <= abs(map->imageData[pos+3] - map->imageData[pos+5]) )
          out->imageData[pos+3]= map->imageData[pos+2];
        else
          out->imageData[pos+3]= map->imageData[pos+5];

        if ( abs(map->imageData[pos+2] - map->imageData[pos+4]) <= abs(map->imageData[pos+4] - map->imageData[pos+5]) )
          out->imageData[pos+4]= map->imageData[pos+2];
        else
          out->imageData[pos+4]= map->imageData[pos+5];
      }
    }
}



bool CDepthInterpolationGeneral::depthsynthesis_3Dwarp( BYTE **pDepthMap, int side ) {
  int h, w, start, end, step;
  int v, u, PREu, ypos;
  bool isLEFT= false;

  CvMat* m = cvCreateMat(3, 1, CV_64F);
  CvMat* mv = cvCreateMat(3, 1, CV_64F);

  cvZero(m_imgVirtualDepth);
  cvZero(m_imgSuccessSynthesis);

  isLEFT= ( side == NONDIR || side == LEFT || side == CENTER );

  if ( isLEFT ) {
    start= 0;
    end= m_uiWidth;
    step= 1;
    PREu= -1;
  } else {
    start= m_uiWidth-1;
    end= 0;
    step= -1;
    PREu= m_uiWidth;
  }

  for ( h=0; h<m_uiHeight; h++ ) {
    for ( w=start; ; w=w+step ) {
      if ( isLEFT && !(w < end) ) break;
      if ( !isLEFT && !(w >= end) ) break;
      cvmSet(m, 0, 0, w);
      cvmSet(m, 1, 0, h);
      cvmSet(m, 2, 0, 1);
      cvmMul(m_matH_R2V[pDepthMap[h][w]], m, mv);
      u= mv->data.db[0] / mv->data.db[2] + 0.5;
      v= mv->data.db[1] / mv->data.db[2] + 0.5;
      // if (u>=0 && u<m_uiWidth && v>=0 && v<m_uiHeight && m_pVirtualDepth[v][u] <= pDepthMap[h][w]) {
      if (u>=0 && u<m_uiWidth && v>=0 && v<m_uiHeight) {
        m_pVirtualDepth[v][u] = pDepthMap[h][w];
        m_pSuccessSynthesis[v][u] = 255;
      }
      // else printf("h: %i, w: %i, v: %i, u: %i PASS\n", h, w, v, u);

      if ( (isLEFT && PREu+1<u) || (!isLEFT && PREu-1>u) ) {
        for ( ypos= PREu;; ypos= ypos+step ) {
          // = は逆かもしれない。
          if ( isLEFT && !(ypos < u) ) break;
          if ( !isLEFT && !(ypos >= u) ) break;
          if ( ypos < 0 ) break;
          if ( ypos >= m_uiWidth ) break;
          if ( m_pSuccessSynthesis[v][ypos] != 255 ) {
            // DEBUG m_pJump[v][ypos] = 255;
            m_pVirtualDepth[v][ypos]= pDepthMap[h][w];
            m_pSuccessSynthesis[v][ypos]= 255;
          }
        }
      }

      PREu= u;
    }
  }

  cvNot(m_imgSuccessSynthesis, m_imgHoles); // m_imgHoles express holes before smoothing

  cvReleaseMat(&m);
  cvReleaseMat(&mv);

  return true;
}



//pSynYuvBuffer->setDataFromDepth(m_pcDepthSynthesisRight->getVirtualDepthMap());
//return true;

bool CDepthInterpolationGeneral::DoOneFrameDepth(
                                  BYTE** RefDepthLeft, BYTE** RefDepthRight, BYTE** RefDepthOutside, CIYuv *pSynYuvBuffer,
                                  double near_C1, double far_C1, double trans_C1,
                                  double near_C2, double far_C2, double trans_C2,
                                  double near_C3, double far_C3, double trans_C3,
                                  double near_V, double far_V, double trans_V,
                                  fn::tri fnum,
                                  unsigned int side, bool mode2view
) {
  BYTE** pRefDepthLeft = RefDepthLeft;
  BYTE** pRefDepthRight = RefDepthRight;
  BYTE** pRefDepthOutside = RefDepthOutside;

  // WARNING: CENTER=QUARTER_ODD/QUARTER_ODD_L mode: C1:LEFT, C2:RIGHT, C3:CENTER !!!!
  if (VariParam) {
    if (
      !m_pcDepthSynthesisLeft->reset_3Dwarp( near_C1, far_C1, trans_C1, near_V, far_V, trans_V )
    ) return false;
    if (
      !m_pcDepthSynthesisRight->reset_3Dwarp( near_C2, far_C2, trans_C2, near_V, far_V, trans_V )
    ) return false;
    if ( !mode2view && side != CENTER )
      if (
        !m_pcDepthSynthesisOutside->reset_3Dwarp( near_C3, far_C3, trans_C3, near_V, far_V, trans_V )
      ) return false;
  }

  if(!m_pcDepthSynthesisLeft->depthsynthesis_3Dwarp(pRefDepthLeft, LEFT)) return false;
  cvSmooth( m_pcDepthSynthesisLeft->getVirtualDepthMap(), m_depthBlended0, CV_MEDIAN, 3 );
  cvSmooth( m_pcDepthSynthesisLeft->getHolePixels(), m_imgLHole, CV_MEDIAN, 3 );
  depthInpaint(m_depthBlended0, m_imgLHole);

  if(!m_pcDepthSynthesisRight->depthsynthesis_3Dwarp(pRefDepthRight, RIGHT)) return false;
  cvSmooth(m_pcDepthSynthesisRight->getVirtualDepthMap(), m_depthBlended1, CV_MEDIAN, 3);
  cvSmooth(m_pcDepthSynthesisRight->getHolePixels(), m_imgRHole, CV_MEDIAN, 3);
  depthInpaint(m_depthBlended1, m_imgRHole);

// DEBUG: Natural Depth //////////////////////////////////
//  char fnameL[20], fnameR[20]; //, fnameL2[20], fnameR2[20], fnameLf[20], fnameRf[20], fnameLj[20], fnameRj[20];
//  int params[] = {CV_IMWRITE_JPEG_QUALITY, 70, -1};
//  sprintf(fnameL, "D3warp%03iL.jpg", fnum.current );
//  sprintf(fnameR, "D3warp%03iR.jpg", fnum.current );
//  sprintf(fnameL2, "D3mf%03iL.jpg", fnum.current);
//  sprintf(fnameR2, "D3mf%03iR.jpg", fnum.current);
//  sprintf(fnameLf, "D3success%03iL.jpg", fnum.current);
//  sprintf(fnameRf, "D3success%03iR.jpg", fnum.current);
//  sprintf(fnameLj, "D3jump%03iL.jpg", fnum.current);
//  sprintf(fnameRj, "D3jump%03iR.jpg", fnum.current);
//  cvSaveImage(fnameL, m_pcDepthSynthesisLeft->getVirtualDepthMap(), params );
//  cvSaveImage(fnameR, m_pcDepthSynthesisRight->getVirtualDepthMap(), params );
//  cvSaveImage(fnameL2, m_depthBlended0, params );
//  cvSaveImage(fnameR2, m_depthBlended1, params );
//  cvSaveImage(fnameLf, m_pcDepthSynthesisLeft->getSynthesizedPixels(), params );
//  cvSaveImage(fnameRf, m_pcDepthSynthesisRight->getSynthesizedPixels(), params );
//  cvSaveImage(fnameLj, m_pcDepthSynthesisLeft->getJumpPixels(), params );
//  cvSaveImage(fnameRj, m_pcDepthSynthesisRight->getJumpPixels(), params );
  ////////////////////////////////////////////////////////////

  if ( side == CENTER ) {
    //m_pcDepthSynthesisOutside->setDepthMap(pRefDepthOutside);
    setDepth( pRefDepthOutside, m_depthBlended2 );
  } else if ( !mode2view ) {
    if(!m_pcDepthSynthesisOutside->depthsynthesis_3Dwarp(pRefDepthOutside, NONDIR)) return false;
    cvSmooth(m_pcDepthSynthesisOutside->getVirtualDepthMap(), m_depthBlended2, CV_MEDIAN, 3);
    cvSmooth(m_pcDepthSynthesisOutside->getHolePixels(), m_imgOHole, CV_MEDIAN, 3);
    depthInpaint(m_depthBlended2, m_imgOHole);
  }


  if ( side == CENTER ) {
    // cvAddWeighted( m_depthBlended0, 0.5, m_depthBlended1, 0.5, 0, m_depthBlendedMix );
    // cvAddWeighted( m_depthBlendedMix, 0.5, m_depthBlended2, 0.5, 0, m_depthBlendedMix2 );
    mf3Depth( m_depthBlended0, m_depthBlended1, m_depthBlended2, m_depthBlendedMix2 );
    depthEnhanced( m_depthBlendedMix2, m_depthBlended0 );
  } else {
    if ( ! mode2view ) {
      // Mar18 cvAddWeighted( m_depthBlended0, 0.5, m_depthBlended1, 0.5, 0, m_depthBlendedMix );
      // Mar18 cvAddWeighted( m_depthBlendedMix, 0.75, m_depthBlended2, 0.25, 0, m_depthBlendedMix2 );
      mf3Depth( m_depthBlended0, m_depthBlended1, m_depthBlended2, m_depthBlendedMix2 );
      depthEnhanced( m_depthBlendedMix2, m_depthBlended0 );
    } else {
      cvAddWeighted( m_depthBlended0, 0.5, m_depthBlended1, 0.5, 0, m_depthBlendedMix2 );
      depthEnhanced( m_depthBlendedMix2, m_depthBlended0 );
    }
  }
  cvSmooth( m_depthBlended0, m_depthBlendedMix, CV_MEDIAN, 5 ); // Sep4,'13 3 -> 5

  pSynYuvBuffer->setDataFromDepth( m_depthBlendedMix ); //m_depthBlended0 ); // m_depthBlendedMix2 ); //m_depthBlendedMix );
  return true;
}



void CDepthInterpolationGeneral::setDepth( BYTE **pDepthMap, IplImage* dst ) {
  for( int h=0, pos=0; h < dst->height; h++ )
    for( int w=0; w < dst->width; w++)
      dst->imageData[pos++]= pDepthMap[h][w];
}



void CDepthInterpolationGeneral::mf3Depth( const IplImage* src1, const IplImage* src2, const IplImage* src3, IplImage* dst ) {
  assert( src1->width == src2->width == src3->width == dst->width );
  assert( src1->height == src2->height == src3->height == dst->height );

  for ( int pos= 0; pos < dst->width * dst->height; pos++ ) {

    if ( (unsigned)src1->imageData[pos] > (unsigned)src2->imageData[pos] ) {
      if ( (unsigned)src1->imageData[pos] > (unsigned)src3->imageData[pos] ) {
        if ( (unsigned)src2->imageData[pos] > (unsigned)src3->imageData[pos] )
          dst->imageData[pos]= src2->imageData[pos];
        else
          dst->imageData[pos]= src3->imageData[pos];
      } else
        dst->imageData[pos]= src1->imageData[pos];
    } else {
      if ( (unsigned)src1->imageData[pos] > (unsigned)src3->imageData[pos] )
        dst->imageData[pos]= src1->imageData[pos];
      else {
        if ( (unsigned)src2->imageData[pos] > (unsigned)src3->imageData[pos] )
          dst->imageData[pos]= src3->imageData[pos];
        else
          dst->imageData[pos]= src2->imageData[pos];
      }
    }
  }
}



// July6,2011 NiCT //////////////////////
void CDepthInterpolationGeneral::ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize)
{
  cvNot(pDepthNEGSRC, m_depthSuccess);
  //cvSmooth(pDepthNEGSRC, pDepthNEGDST, CV_GAUSSIAN , fsize, 3); // CV_MEDIAN, fsize);
  cvSmooth(pDepthNEGSRC, pDepthNEGDST, CV_MEDIAN, fsize);
  cvAnd(m_depthSuccess, pDepthNEGDST, m_imgMask[4]);
  cvCopy(pDepthNEGSRC, pDepthNEGDST, m_imgMask[4]);

  //cvSmooth(pDepthSRC, pDepthDST, CV_GAUSSIAN , fsize, 3); // CV_MEDIAN, fsize);
  cvSmooth(pDepthSRC, pDepthDST, CV_MEDIAN, fsize);
  cvCopy(pDepthSRC, pDepthDST, m_imgMask[4]);
}
