// View Synthesis Encoder for Residuals
// Dev Enviroment:
//    Windows 8.1 64bit with Visual Studio 2013(VS12) C++ x64
//    OpenCV 2.4.9 64bit
// Original code: View Synthesis Reference Software (VSRS) 3.5
// Designer: Senoh, NiCT, Tokyo. Japan.
// Last Update: June, 2014.

#include "SubMixOptions.h"
#include "ResidualSynthesis.h"
#include <string.h>
#include <stdlib.h>
#include <opencv2/legacy/compat.hpp> // for cvmMul()
#include <opencv2/calib3d/calib3d.hpp> // for cvFindHomography()
#include <opencv2/imgproc/imgproc_c.h> // for cvSmooth(), cvDilate()
// #include <opencv2/photo/photo_c.h> // for cvInpaint()
// #include <opencv2/highgui/highgui_c.h> // for DEBUG cvSaveImage()

#ifndef SAFE_RELEASE_IMAGE
#define SAFE_RELEASE_IMAGE(p) { if((p)!=NULL){ cvReleaseImage(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_RELEASE_MAT
#define SAFE_RELEASE_MAT(p) { if((p)!=NULL){ cvReleaseMat(&(p)); (p)=NULL; } }
#endif
#ifndef SAFE_FREE
#define SAFE_FREE(p) { if((p)!=NULL){ free(p); (p)=NULL; } }
#endif

#define STANDARDWIDTH 1920
#define DILATE 1

#define NARROW 0
#define WIDE 1

CResidualsInterpolation::CResidualsInterpolation() {
  int i;

  m_uiPrecision= 1;

  m_imgVirtualImage= m_imgVirtualDepth= m_imgWorkDepth= NULL;
  m_imgSuccessSynthesis = m_imgHoles = NULL;
  m_pVirtualDepth= m_pWorkDepth= m_pSuccessSynthesis= m_pHole= m_pHole1= NULL;
  m_depthS1= m_depthS2= m_depthS3= m_depthS4= m_depthR1= m_depthL1= m_depthL1a= NULL;

  m_pS1= m_pS1a= m_pS2= m_pS3= m_pS2d= m_pS3d= NULL;
  m_pL1= m_pL1a= m_pL2= m_pL3= m_pR1= m_pR1a= m_pD1= NULL;

  m_imgInterpolatedView= NULL; // =m_imgInterpolatedView2
  m_depthSuccess= NULL;

  for(i=0; i<5; i++)
    m_imgTemp[i] = m_imgMask[i] = NULL;

  for(i=0; i<256; i++)
    m_matH_R2V[i] = m_matH_V2R[i] = NULL;

  // for VariParam
  for(i=0; i<2; i++)
    m_Mat_in[i]= m_Mat_ex_c2w[i]= m_Mat_proj_w2i[i]= NULL;

  cv::setUseOptimized(true);
  //if (cv::useOptimized()) printf("\nuseOptimized(): true"); else printf("\nuseOptimized(): false");
}



CResidualsInterpolation::~CResidualsInterpolation() {
  xReleaseMemory();
}



void CResidualsInterpolation::xReleaseMemory() {
  int i;

  SAFE_RELEASE_IMAGE(m_imgVirtualImage);
  SAFE_RELEASE_IMAGE(m_imgVirtualDepth);
  SAFE_RELEASE_IMAGE(m_imgWorkDepth);
  SAFE_RELEASE_IMAGE(m_imgSuccessSynthesis);
  SAFE_RELEASE_IMAGE(m_imgHoles);
  SAFE_RELEASE_IMAGE(m_imgInterpolatedView);
  //SAFE_RELEASE_IMAGE(m_imgInterpolatedView2);
  SAFE_RELEASE_IMAGE(m_depthSuccess); // Jul4,2011 NiCT
  SAFE_RELEASE_IMAGE(m_depthS1);
  SAFE_RELEASE_IMAGE(m_depthS2);
  SAFE_RELEASE_IMAGE(m_depthS3);
  SAFE_RELEASE_IMAGE(m_depthS4);
  SAFE_RELEASE_IMAGE(m_depthR1);
  SAFE_RELEASE_IMAGE(m_depthL1);
  SAFE_RELEASE_IMAGE(m_depthL1a);

  if(m_pS1!=NULL) delete m_pS1;
  if(m_pS1a!=NULL) delete m_pS1a;
  if(m_pS2!=NULL) delete m_pS2;
  if(m_pS3!=NULL) delete m_pS3;
  if(m_pS2d!=NULL) delete m_pS2d;
  if(m_pS3d!=NULL) delete m_pS3d;
  if(m_pL1!=NULL) delete m_pL1;
  if(m_pL1a!=NULL) delete m_pL1a;
  if(m_pL2!=NULL) delete m_pL2;
  if(m_pL3!=NULL) delete m_pL3;
  if(m_pR1!=NULL) delete m_pR1;
  if(m_pR1a!=NULL) delete m_pR1a;
  if(m_pD1!=NULL) delete m_pD1;
  m_pS1= m_pS1a= m_pS2= m_pS3= m_pS2d= m_pS3d= NULL;
  m_pL1= m_pL1a= m_pL2= m_pL3= m_pR1= m_pR1a= m_pD1= NULL;

  for (i=0; i<5; i++) {
    SAFE_RELEASE_IMAGE(m_imgTemp[i]);
    SAFE_RELEASE_IMAGE(m_imgMask[i]);
  }

  SAFE_FREE(m_pVirtualDepth);
  SAFE_FREE(m_pWorkDepth);
  SAFE_FREE(m_pSuccessSynthesis);
  SAFE_FREE(m_pHole);
  SAFE_FREE(m_pHole1);

  for (i=0; i<256; i++) {
    SAFE_RELEASE_MAT( m_matH_R2V[i] );
    SAFE_RELEASE_MAT( m_matH_V2R[i] );
  }

  // for VariParam
  for(i=0; i<2; i++) {
    SAFE_RELEASE_MAT( m_Mat_in[i] );
    SAFE_RELEASE_MAT( m_Mat_ex_c2w[i] );
    SAFE_RELEASE_MAT( m_Mat_proj_w2i[i] );
  }
}



void CResidualsInterpolation::convertCameraParam(CvMat *exMat_dst, CvMat *exMat_src)
{
  int i, j;
  double val;

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      cvmSet(exMat_dst, i, j, cvmGet(exMat_src, j, i));

  for (i=0; i<3; i++) {
    val = 0.0;
    for(j=0; j<3; j++)
      val -= cvmGet(exMat_dst, i, j)*cvmGet(exMat_src, j, 3);
    cvmSet(exMat_dst, i, 3, val);
  }
}



// Winter Versionでは、Call側でカメラ位置を左右反転させて指定する。
// LEFT:  C1, C2, V1, A1, A2
// RIGHT: C3, C2, V2, A4, A3
bool CResidualsInterpolation::InitOneSide (
    unsigned int width, unsigned int height, bool variparam,
    double near1, double far1, double near2, double far2,
    const char *RefCam1, const char *RefCam2, const char *VirCam1, const char *AddCam1, const char *AddCam2,
    const unsigned int Side, const unsigned int mode,
    double Mat_In_1[9], double Mat_Ex_1[9], double Mat_Trans_1[3],
    double Mat_In_2[9], double Mat_Ex_2[9], double Mat_Trans_2[3],
    double Mat_In_V1[9], double Mat_Ex_V1[9], double Mat_Trans_V1[3],
    double Mat_In_A1[9], double Mat_Ex_A1[9], double Mat_Trans_A1[3],
    double Mat_In_A2[9], double Mat_Ex_A2[9], double Mat_Trans_A2[3]
) {
  VariParam= variparam;

  m_ResidualsSide= Side;
  m_uiWidth= width;
  m_uiHeight= height;

  //// SHIFT ch /////////////////////////////////////////////////////////
  if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) { //////////////////////////////////////////

    m_pD1= new CResidualsInterpolation();
    if (m_pD1 == NULL) return false;
    if (!m_pD1->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

    m_pS1= new CResidualsInterpolation();
    if (m_pS1 == NULL) return false;
    if (!m_pS1->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

    m_pS2= new CResidualsInterpolation();
    if (m_pS2 == NULL) return false;
    if (!m_pS2->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_A1, Mat_Ex_A1, Mat_Trans_A1
    ) ) return false;

    m_pS3= new CResidualsInterpolation();
    if (m_pS3 == NULL) return false;
    if (!m_pS3->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_A2, Mat_Ex_A2, Mat_Trans_A2
    ) ) return false;

    m_pS1a= new CResidualsInterpolation();
    if (m_pS1a == NULL) return false;
    if (!m_pS1a->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

    m_pR1a= new CResidualsInterpolation();
    if (m_pR1a == NULL) return false;
    if (!m_pR1a->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1
    ) ) return false;

    m_pL1a= new CResidualsInterpolation();
    if (m_pL1a == NULL) return false;
    if (!m_pL1a->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_2, Mat_Ex_2, Mat_Trans_2,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

  } else { ///////////////////////////////////////////////////////////////

    m_pD1= new CResidualsInterpolation();
    if (m_pD1 == NULL) return false;
    if (!m_pD1->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

    m_pS1= new CResidualsInterpolation();
    if (m_pS1 == NULL) return false;
    if (!m_pS1->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                Mat_In_1, Mat_Ex_1, Mat_Trans_1
    ) ) return false;

    m_pS2= new CResidualsInterpolation();
    if (m_pS2 == NULL) return false;
    if (!m_pS2->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                Mat_In_A1, Mat_Ex_A1, Mat_Trans_A1
    ) ) return false;

    m_pS3= new CResidualsInterpolation();
    if (m_pS3 == NULL) return false;
    if (!m_pS3->Init(
                width, height, variparam,
                near1, far1, Side,
                Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1,
                Mat_In_A2, Mat_Ex_A2, Mat_Trans_A2
    ) ) return false;

  }

  m_pS2d= new CResidualsInterpolation();
  if (m_pS2d == NULL) return false;
  if (!m_pS2d->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_A1, Mat_Ex_A1, Mat_Trans_A1,
              Mat_In_1, Mat_Ex_1, Mat_Trans_1
  ) ) return false;

  m_pS3d= new CResidualsInterpolation();
  if (m_pS3d == NULL) return false;
  if (!m_pS3d->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_A2, Mat_Ex_A2, Mat_Trans_A2,
              Mat_In_1, Mat_Ex_1, Mat_Trans_1
  ) ) return false;

  //// Left ch ////////////////////////////////////////////////////////
  m_pL1= new CResidualsInterpolation();
  if (m_pL1 == NULL) return false;
  if (!m_pL1->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_2, Mat_Ex_2, Mat_Trans_2,
              Mat_In_1, Mat_Ex_1, Mat_Trans_1
  ) ) return false;

  m_pL2= new CResidualsInterpolation();
  if (m_pL2 == NULL) return false;
  if (!m_pL2->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_2, Mat_Ex_2, Mat_Trans_2,
              Mat_In_A1, Mat_Ex_A1, Mat_Trans_A1
  ) ) return false;

  m_pL3= new CResidualsInterpolation();
  if (m_pL3 == NULL) return false;
  if (!m_pL3->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_2, Mat_Ex_2, Mat_Trans_2,
              Mat_In_A2, Mat_Ex_A2, Mat_Trans_A2
  ) ) return false;

  //// RIGHT ch ////////////////////////////////////////////////////////
  m_pR1= new CResidualsInterpolation();
  if (m_pR1 == NULL) return false;
  if (!m_pR1->Init(
              width, height, variparam,
              near1, far1, Side,
              Mat_In_2, Mat_Ex_2, Mat_Trans_2,
              Mat_In_V1, Mat_Ex_V1, Mat_Trans_V1
  ) ) return false;

/*
  double dTotalBaseline;
  m_dWeightLeft = m_pcViewHoleRight->getBaselineDistance();
  m_dWeightRight = m_pcViewHoleLeft->getBaselineDistance();
  dTotalBaseline = m_dWeightLeft + m_dWeightRight;
#ifdef FILOUT
  printf("\nCamera's base line  L:%f, R:%f, LR:%f.\n", m_dWeightLeft, m_dWeightRight, dTotalBaseline);
#endif
  m_dWeightLeft /= dTotalBaseline;
  m_dWeightRight /= dTotalBaseline;
#ifdef FILOUT
  printf("Camera's weight: L:%f, R:%f.\n", m_dWeightLeft, m_dWeightRight);
#endif
*/

  m_depthS1= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthS2= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthS3= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthS4= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthR1= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthL1= cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthL1a= cvCreateImage(cvSize(Width, Height), 8, 1);

  m_imgInterpolatedView = cvCreateImage(cvSize(Width, Height), 8, 3);
  //m_imgInterpolatedView2 = cvCreateImage(cvSize(Width, Height), 8, 3);
  m_imgMask[2]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgMask[3]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgMask[4]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgTemp[0]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgTemp[1]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgTemp[2]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgTemp[3]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_imgTemp[4]  = cvCreateImage(cvSize(Width, Height), 8, 1);
  m_depthSuccess= cvCreateImage(cvSize(Width, Height), 8, 1);

  return true;
}



bool CResidualsInterpolation::Init(
              unsigned int width, unsigned int height, bool variparam,
              double dNear, double dFar, unsigned int Side,
              double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
              double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3]
) {
  unsigned int h, pos;
  double fB;

  xReleaseMemory();

  m_ResidualsSide= Side;
  m_uiHeight= height;
  m_uiWidth= width;

  fB = fabs(Mat_In_Ref[0]) * fabs(Mat_Trans_Ref[0] - Mat_Trans_Vir[0]);
  m_dMaxDisp = fB / fabs(dNear);
  m_dispA = 1.0 / 255.0 * fB * fabs(1.0 / dNear - 1.0 / dFar);
  m_dispB = fB / fabs(dFar);
  //printf("\nfB: %f, m_dMaxDisp: %f, m_dispA: %f, m_dispB: %f", fB, m_dMaxDisp, m_dispA, m_dispB);

  m_imgVirtualImage = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 3);
  m_imgVirtualDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgWorkDepth = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgSuccessSynthesis = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgHoles    = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgMask[0]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);
  m_imgMask[1]  = cvCreateImage(cvSize(m_uiWidth, m_uiHeight), 8, 1);

  if((m_pVirtualDepth = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgVirtualDepth->widthStep)
    m_pVirtualDepth[h] = (BYTE *) &(m_imgVirtualDepth->imageData[pos]);

  if((m_pWorkDepth = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgWorkDepth->widthStep)
    m_pWorkDepth[h] = (BYTE *) &(m_imgWorkDepth->imageData[pos]);

  if((m_pSuccessSynthesis = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgSuccessSynthesis->widthStep)
    m_pSuccessSynthesis[h] = (BYTE *) &(m_imgSuccessSynthesis->imageData[pos]);

  if((m_pHole = (BYTE **)malloc(m_uiHeight*sizeof(BYTE *)))==NULL) return false;
  for(h=pos=0; h<m_uiHeight; h++, pos+=m_imgHoles->widthStep)
    m_pHole[h] = (BYTE *) &(m_imgHoles->imageData[pos]);

//  if (VariParam) {
  for(unsigned int i=0; i<2; i++) {
    m_Mat_in[i] = cvCreateMat(3, 3, CV_64F); //intrinsic parameter of camera (3x3 matrix)
    m_Mat_ex_c2w[i] = cvCreateMat(3, 4, CV_64F); //extrinsic parameter of camera (3x4 matrix)
    m_Mat_proj_w2i[i] = cvCreateMat(3, 4, CV_64F); // projection matrix
  }

  for (unsigned int i=0; i<9; i++) Mat_In_C[i]= Mat_In_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_C[i]= Mat_Ex_Ref[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_C[i]= Mat_Trans_Ref[i];
  for (unsigned int i=0; i<9; i++) Mat_In_V[i]= Mat_In_Vir[i];
  for (unsigned int i=0; i<9; i++) Mat_Ex_V[i]= Mat_Ex_Vir[i];
  for (unsigned int i=0; i<3; i++) Mat_Trans_V[i]= Mat_Trans_Vir[i];
//  }

  return init_3Dwarp( dNear, dFar );

}



bool CResidualsInterpolation::init_camera_param(
                                double Mat_In_Ref[9], double Mat_Ex_Ref[9], double Mat_Trans_Ref[3],
                                double Mat_In_Vir[9], double Mat_Ex_Vir[9], double Mat_Trans_Vir[3],
                                CvMat *mat_in[2], CvMat *mat_ex_c2w[2], CvMat *mat_proj[2]
) {
  int i, j;
  CvMat *mat_ex_w2c;

  for(i=0; i<3; i++)
  {
    for(j=0; j<3;j++)
    {
      mat_in[0]->data.db[j+3*i] = *(Mat_In_Ref+3*i+j);
      mat_in[1]->data.db[j+3*i] = *(Mat_In_Vir+3*i+j);
    }
  }
  for(i=0; i<3; i++)
  {
    for(j=0; j<3;j++)
    {
      mat_ex_c2w[0]->data.db[j+4*i] = *(Mat_Ex_Ref+3*i+j);
      mat_ex_c2w[1]->data.db[j+4*i] = *(Mat_Ex_Vir+3*i+j);
    }
  }
  for(i=0; i<3; i++)
  {
    mat_ex_c2w[0]->data.db[3+4*i] = *(Mat_Trans_Ref+i);
    mat_ex_c2w[1]->data.db[3+4*i] = *(Mat_Trans_Vir+i);
  }

  if(mat_proj!=NULL)
  {
    mat_ex_w2c = cvCreateMat(3, 4, CV_64F);
    for(i=0; i<2; i++)
    {
      convertCameraParam(mat_ex_w2c, mat_ex_c2w[i]);
      cvmMul(mat_in[i], mat_ex_w2c, mat_proj[i]);  // Proj = inMat_c2i * exMat_w2c
    }
    cvReleaseMat(&mat_ex_w2c);
  }

  return true;
}



void CResidualsInterpolation::image2world_with_z(
                              CvMat *mat_Rc2w_invIN_from, CvMat *matEX_c2w_from, CvMat *image, CvMat *world
) {
  CvMat *temp;
  double s;

  temp= cvCreateMat(3, 1, CV_64F);

  cvmMul(mat_Rc2w_invIN_from, image, temp);

  s= ( cvmGet(world, 2, 0) - cvmGet(matEX_c2w_from, 2, 3) ) / cvmGet(temp, 2, 0);

  cvmSet(world, 0, 0, s*cvmGet(temp, 0, 0) + cvmGet(matEX_c2w_from, 0, 3));
  cvmSet(world, 1, 0, s*cvmGet(temp, 1, 0) + cvmGet(matEX_c2w_from, 1, 3));

  cvReleaseMat(&temp);
}



void CResidualsInterpolation::makeHomography(
                              CvMat *matH_F2T[256], CvMat *matH_T2F[256],
                              double adTable[256],
                              CvMat *matIN_from, CvMat *matEX_c2w_from, CvMat *matProj_to,
                              unsigned int work
) {
  int i, j, k;
  double val, u, v;
  CvMat *matIN_inv_from, *mat_Rc2w_invIN_from;

  CvMat *src_points  = cvCreateMat(4, 2, CV_64F);
  CvMat *dst_points  = cvCreateMat(4, 2, CV_64F);
  CvMat *image      = cvCreateMat(3, 1, CV_64F);
  CvMat *world      = cvCreateMat(4, 1, CV_64F);

  matIN_inv_from      = cvCreateMat(3, 3, CV_64F);
  mat_Rc2w_invIN_from  = cvCreateMat(3, 3, CV_64F);

  cvInvert(matIN_from, matIN_inv_from, CV_LU);
  for(i=0; i<3; i++) {
    for(j=0; j<3; j++) {
      val=0.0;
      for(k=0; k<3; k++)
      {
        val += cvmGet(matEX_c2w_from, i, k)*cvmGet(matIN_inv_from, k, j);
      }
      cvmSet(mat_Rc2w_invIN_from, i, j, val);
    }
  }

  if ( work == INIT )
    for (unsigned int i=0; i<256; i++) {
      matH_F2T[i]= cvCreateMat(3, 3, CV_64F);
      matH_T2F[i]= cvCreateMat(3, 3, CV_64F);
    }

  for(i=0; i<256; i++)
  {
    cvmSet(world, 2, 0, adTable[i]);
    cvmSet(world, 3, 0, 1.0);

    // (u, v) = (0, 0)
    u = 0.0; v= 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 0, 0, u);
    cvmSet(src_points, 0, 1, v);
    cvmSet(dst_points, 0, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 0, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (0, height-1)
    u = 0.0; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 1, 0, u);
    cvmSet(src_points, 1, 1, v);
    cvmSet(dst_points, 1, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 1, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, height-1)
    u = m_uiWidth - 1; v = m_uiHeight - 1;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 2, 0, u);
    cvmSet(src_points, 2, 1, v);
    cvmSet(dst_points, 2, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 2, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    // (u, v) = (width-1, 0)
    u = m_uiWidth - 1; v = 0.0;
    cvmSet(image, 0, 0, u);
    cvmSet(image, 1, 0, v);
    cvmSet(image, 2, 0, 1.0);
    image2world_with_z(mat_Rc2w_invIN_from, matEX_c2w_from, image, world);
    cvmMul(matProj_to, world, image);
    cvmSet(src_points, 3, 0, u);
    cvmSet(src_points, 3, 1, v);
    cvmSet(dst_points, 3, 0, cvmGet(image, 0, 0) / cvmGet(image, 2, 0));
    cvmSet(dst_points, 3, 1, cvmGet(image, 1, 0) / cvmGet(image, 2, 0));

    cvFindHomography(src_points, dst_points, matH_F2T[i]);
    cvInvert(matH_F2T[i], matH_T2F[i], CV_LU);
  }

  cvReleaseMat(&src_points);
  cvReleaseMat(&dst_points);
  cvReleaseMat(&image);
  cvReleaseMat(&world);
  cvReleaseMat(&matIN_inv_from);
  cvReleaseMat(&mat_Rc2w_invIN_from);
}



bool CResidualsInterpolation::init_3Dwarp( double Z_near, double Z_far ) {
  double tableD2Z[256];
  double distance, temp1, temp2;

  if( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V, m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1= 1.0/Z_near - 1.0/Z_far;
  temp2= 1.0/Z_far;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  makeHomography( m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1], INIT );

  return true;
  //////////////////////////////////////////////////////////////
  m_dBaselineDistance=  pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance= sqrt(m_dBaselineDistance);
  return true;
}



bool CResidualsInterpolation::reset_3Dwarp(
                                    double near_C, double far_C, double trans_C,
                                    double near_V, double far_V, double trans_V
) {
  double tableD2Z[256];
  double distance, temp1, temp2;
  double fB;

  Mat_Trans_C[0]= trans_C;
  Mat_Trans_V[0]= trans_V;
  if( !init_camera_param(
          Mat_In_C, Mat_Ex_C, Mat_Trans_C,
          Mat_In_V, Mat_Ex_V, Mat_Trans_V,
          m_Mat_in, m_Mat_ex_c2w, m_Mat_proj_w2i)
  ) return false;

  temp1= 1.0 / near_C - 1.0 / far_C;
  temp2= 1.0 / far_C;

  // uiDepthType == 0
  for(unsigned int i= 0; i<256; i++) {
    distance = 1.0 / ( double(i)*temp1/255.0 + temp2 );
    tableD2Z[i] = cvmGet(m_Mat_ex_c2w[0], 2, 2) * distance + cvmGet(m_Mat_ex_c2w[0], 2, 3);
  }
  // uiDepthType == 1
  //for(unsigned i=0; i<256; i++) tableD2Z[i] = 1.0 / ( double(i)*temp1/255.0 + temp2 );

  makeHomography( m_matH_R2V, m_matH_V2R, tableD2Z, m_Mat_in[0], m_Mat_ex_c2w[0], m_Mat_proj_w2i[1] );

  fB = fabs(Mat_In_C[0]) * fabs(Mat_Trans_C[0] - Mat_Trans_V[0]);
  m_dMaxDisp = fB / fabs(near_C);
  m_dispA = 1.0 / 255.0 * fB * fabs(1.0 / near_C - 1.0 / far_C);
  m_dispB = fB / fabs(far_C);

  return true;
  //////////////////////////////////////////////////////////////
  m_dBaselineDistance=  pow( (m_Mat_ex_c2w[0]->data.db[ 3] - m_Mat_ex_c2w[1]->data.db[ 3]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[ 7] - m_Mat_ex_c2w[1]->data.db[ 7]), 2.0 ) +
                        pow( (m_Mat_ex_c2w[0]->data.db[11] - m_Mat_ex_c2w[1]->data.db[11]), 2.0 );
  m_dBaselineDistance= sqrt(m_dBaselineDistance);

  return true;
}



bool CResidualsInterpolation::setDepth(BYTE **pDepthMap)
{
  unsigned int h, w;

  for(h=0; h<m_uiHeight; h++)
    for(w=0; w<m_uiWidth; w++) {
      if (pDepthMap != NULL) m_pVirtualDepth[h][w]= pDepthMap[h][w];
      m_pSuccessSynthesis[h][w] = 255;
    }

  return true;
}



bool CResidualsInterpolation::setView(BYTE ***src)
{
  unsigned int h, w, ptv;

  for(h=0; h<m_uiHeight; h++)
    for(w=0; w<m_uiWidth; w++) {
      ptv = w + h * m_uiWidth;
      m_imgVirtualImage->imageData[ptv * 3     ] = src[0][h][w];
      m_imgVirtualImage->imageData[ptv * 3 + 1 ] = src[1][h][w];
      m_imgVirtualImage->imageData[ptv * 3 + 2 ] = src[2][h][w];
      m_pSuccessSynthesis[h][w] = 255;
    }

  return true;
}



void CResidualsInterpolation::depthInpaint( IplImage* map, IplImage* hole ) {
  int pos= 0;
  bool area= false;
  bool work= false;
  int start_w, stop_w, paint_pos; //, back_w, backpos;
  unsigned int left_value, right_value, paint_value;

  //return;

  for ( int h= 0; h < map->height; h++ ) {
    area= false;
    work= false;
    for ( int w= 0; w < map->width; w++, pos++ ) {
      if ( !area && hole->imageData[pos] == -1 ) { // 255
        area= true;
        start_w= w;
        if ( start_w == 0 )
          left_value= 255; // 255; -1
        else
          left_value= (unsigned char)map->imageData[pos-1];
      } else if ( area && hole->imageData[pos] == 0 ) {
        area= false;
        work= true;
        stop_w= w - 1;
        right_value= (unsigned char)map->imageData[pos];
      }
      if ( area && (w == (map->width - 1)) ) {
        work= true;
        area= false;
        stop_w= w;
        right_value= 255;
      }

      if ( work ) {
        if ( left_value <= right_value )
          paint_value= left_value;
        else
          paint_value= right_value;

        for ( int paint_w= start_w; paint_w <= stop_w; paint_w++ ) {
          paint_pos= map->width * h + paint_w;
          //DEBUG printf(", %i", paint_pos);
          map->imageData[paint_pos]= paint_value; // (char) ?
          hole->imageData[paint_pos]= 0;
        }
        work= false;
        area= false;
      }
    }
  }
}



// Shift for Depth & Hole image.
bool CResidualsInterpolation::depthShift(BYTE **pDepthMap, unsigned int from, int MODE)
{
  int h, w, v, u, PREu, ypos, start, end, step;
  bool isLEFT= false;

  CvMat* m = cvCreateMat(3, 1, CV_64F);
  CvMat* mv = cvCreateMat(3, 1, CV_64F);

  cvZero(m_imgVirtualDepth);
  cvZero(m_imgSuccessSynthesis);

  isLEFT= ( from == LEFT );
  if ( isLEFT ) {
    start= 0;
    end= m_uiWidth;
    step= 1;
    PREu= -1;
  } else {
    start= m_uiWidth-1;
    end= 0;
    step= -1;
    PREu= m_uiWidth;
  }

  for ( h=0; h<m_uiHeight; h++ ) {
    for ( w=start; ; w=w+step ) {
      if ( isLEFT && !(w < end) ) break;
      if ( !isLEFT && !(w >= end) ) break;

      cvmSet(m, 0, 0, w);
      cvmSet(m, 1, 0, h);
      cvmSet(m, 2, 0, 1);

      cvmMul( m_matH_R2V[ pDepthMap[h][w] ], m, mv );
      u = mv->data.db[0] / mv->data.db[2] + 0.5;
      v = mv->data.db[1] / mv->data.db[2] + 0.5;

      if ( u>=0 && u<m_uiWidth && v>=0 && v<m_uiHeight ) { // && m_pVirtualDepth[v][u] <= pDepthMap[h][w]
        if ( MODE == NORMAL ) {
          m_pVirtualDepth[v][u] = pDepthMap[h][w];
          m_pSuccessSynthesis[v][u] = 255;
        } else { // HOLESHIFT
          if ( m_pVirtualDepth[v][u] == 0 ) {
            m_pVirtualDepth[v][u] = m_pHole[h][w];
            m_pSuccessSynthesis[v][u] = 255;
          }
        }
      }

      if ( ( isLEFT && PREu+1<u ) || ( !isLEFT && PREu-1>u ) ) {
        for ( ypos= PREu;; ypos= ypos+step ) {
          // = は逆かもしれない。
          if ( isLEFT && !(ypos < u) ) break;
          if ( !isLEFT && !(ypos >= u) ) break;
          if ( ypos < 0 ) break;
          if ( ypos >= m_uiWidth ) break;
          if ( m_pSuccessSynthesis[v][ypos] != 255 ) {
            // DEBUG m_pJump[v][ypos] = 255;
            if ( MODE == NORMAL ) {
              m_pVirtualDepth[v][ypos]= pDepthMap[h][w];
              m_pSuccessSynthesis[v][ypos]= 255;
            } else { // HOLESHIFT
              //m_pVirtualDepth[v][ypos] = m_pHole[h][w];
              //m_pSuccessSynthesis[v][ypos] = 255;
            }
          }
        }
      }

      PREu= u;
    }
  }
  cvNot(m_imgSuccessSynthesis, m_imgHoles);

  cvReleaseMat( &m );
  cvReleaseMat( &mv );
  return true;
}



#define DEPTHGAP 5 // 4   6
#define DEPTHALLOW 0.4 // 0.5   0.7
bool CResidualsInterpolation::depthsynthesis_reverse_3Dwarp(BYTE **pDepthMap)
{
  int u, v, h, w, PREu, ypos;
  int start, end, step;
  bool isLEFT= false;

  double c, d, disp, dMaxDisp, dMaxStep;
  int si, ww;

  CvMat* m= cvCreateMat(3, 1, CV_64F);
  CvMat* mv= cvCreateMat(3, 1, CV_64F);

  cvZero(m_imgVirtualImage);
  cvZero(m_imgHoles);

  isLEFT= ( m_ResidualsSide == LEFT );
  if ( isLEFT ) {
    start= 0;
    end= m_uiWidth;
    step= 1;
    PREu= 0;
  } else {
    start= m_uiWidth-1;
    end= 0;
    step= -1;
    PREu= m_uiWidth;
  }

  for ( h=0; h<(int)m_uiHeight; h++ ) {
    for ( w=start; ; w=w+step ) {
      if ( isLEFT && !(w < end) ) break;
      if ( !isLEFT && !(w >= end) ) break;

      dMaxDisp= m_dMaxDisp;
      do {
        if ( isLEFT ) dMaxStep= dMaxDisp; else dMaxStep= -dMaxDisp;
        si= w + dMaxStep; // 近傍画素位置一
        if ( si >= 0 && si < (int)m_uiWidth ) {
          c= m_pVirtualDepth[h][si] - m_pVirtualDepth[h][w]; // 近傍画素位置１と現在位置のデプス値差分
          disp= m_dispA * c + m_dispB;
          if ( c > DEPTHGAP ) {
            if ( disp > dMaxDisp ) {
              m_pHole[h][w]= 255;
              break; /////////////////////////////////
            }
            if ( isLEFT ) ww= w + disp; else ww= w - disp; // 近傍画素位置２＝現在位置＋デプス差分
            if ( ww >= 0 && ww < (int)m_uiWidth ) {
              d= m_pVirtualDepth[h][ww] - m_pVirtualDepth[h][w]; // 近傍画素位置２と現在位置のデプス値差分
              c*= DEPTHALLOW; // 近傍画素位置１と現在位置のデプス値差分をマイルドに
              if ( d >= c ) {
                m_pHole[h][w]= 255;
                break; /////////////////////////////////
              }
            }
          }
        }

        dMaxDisp-= m_dMaxDisp/8;
      } while( dMaxDisp > 0 );


      if ( m_pHole[h][w] == 0 ) {
        cvmSet( m, 0, 0, w );
        cvmSet( m, 1, 0, h );
        cvmSet( m, 2, 0, 1 );
        cvmMul( m_matH_V2R[m_pVirtualDepth[h][w]], m, mv );
        u = mv->data.db[0] / mv->data.db[2] + 0.5;
        v = mv->data.db[1] / mv->data.db[2] + 0.5;

        if( u<0 || u>=(int)m_uiWidth || v<0 || v>=(int)m_uiHeight ) m_pHole[h][w]= 255;
      }

    }
  }

  cvReleaseMat(&m);
  cvReleaseMat(&mv);

  return true;
}



int CResidualsInterpolation::DoOneFrame ( unsigned int fnum ) {
  unsigned int h, pos, from;

  if (VariParam) {
    if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) {                                                 // case of side LEFT
      if( !m_pS1->reset_3Dwarp( near_C2, far_C2, trans_C2, near_C1, far_C1, trans_C1 ) ) return false;    // from R to L
      if( !m_pS2->reset_3Dwarp( near_C2, far_C2, trans_C2, near_A1, far_A1, trans_A1 ) ) return false;    // from R
      if( !m_pS3->reset_3Dwarp( near_C2, far_C2, trans_C2, near_A2, far_A2, trans_A2 ) ) return false;    // from R
      if( !m_pS1a->reset_3Dwarp( near_V1, far_V1, trans_V1, near_C1, far_C1, trans_C1 ) ) return false;   // from R
      if( !m_pR1a->reset_3Dwarp( near_C2, far_C2, trans_C2, near_V1, far_V1, trans_V1 ) ) return false;   // from R
      if( !m_pL1a->reset_3Dwarp( near_C2, far_C2, trans_C2, near_C1, far_C1, trans_C1 ) ) return false;   // from R
    } else {
      if( !m_pS1->reset_3Dwarp( near_V1, far_V1, trans_V1, near_C1, far_C1, trans_C1 ) ) return false;    // from R
      if( !m_pS2->reset_3Dwarp( near_V1, far_V1, trans_V1, near_A1, far_A1, trans_A1 ) ) return false;    // from R
      if( !m_pS3->reset_3Dwarp( near_V1, far_V1, trans_V1, near_A2, far_A2, trans_A2 ) ) return false;    // from L to R
    }
    if( !m_pS2d->reset_3Dwarp( near_A1, far_A1, trans_A1, near_C1, far_C1, trans_C1 ) ) return false;     // from R
    if( !m_pS3d->reset_3Dwarp( near_A2, far_A2, trans_A2, near_C1, far_C1, trans_C1 ) ) return false;     // from R
    if( !m_pL1->reset_3Dwarp( near_C2, far_C2, trans_C2, near_C1, far_C1, trans_C1 ) ) return false;      // from R
    if( !m_pL2->reset_3Dwarp( near_C2, far_C2, trans_C2, near_A1, far_A1, trans_A1 ) ) return false;      // from R
    if( !m_pL3->reset_3Dwarp( near_C2, far_C2, trans_C2, near_A2, far_A2, trans_A2 ) ) return false;      // from R
    if( !m_pR1->reset_3Dwarp( near_C2, far_C2, trans_C2, near_V1, far_V1, trans_V1 ) ) return false; // L4// from R
    printf(" V ");
  }

  // DEBUG: 素材無加工Depth
//  m_pS1->setDepth(Depth);
//  cvSaveImage(fname, m_pS1->getVirtualDepthMap());
//  return true;

  //cvSmooth(Depth, m_pD1->getWorkDepthMap(), CV_MEDIAN, 3);
  m_pD1->setDepth(Depth);
  cvSmooth(m_pD1->getVirtualDepthMap(), m_pD1->getWorkDepthMap(), CV_MEDIAN, 5);

  // C1:Step.1 L:V1 -> C1, R:V2 -> C3 ///////////////////////////////////////////////
  //          Vertual Depth を Residuals側までシフトする。
  //          そのデプスをMain側に投影し、穴を求める。
  if ( side == LEFT ) from= RIGHT; else from= LEFT;
  if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) { //////////////////////////////////////////////////////
    if(!m_pL1a->depthShift(m_pD1->m_pWorkDepth, from)) return false;
    cvSmooth(m_pL1a->getVirtualDepthMap(), m_depthL1a, CV_MEDIAN, 3);
    cvSmooth(m_pL1a->getHolePixels(), m_imgMask[2], CV_MEDIAN, 3);
    depthInpaint(m_depthL1a, m_imgMask[2]);

    for(h=pos=0; h<m_pL1->m_uiHeight; h++, pos+=m_depthL1a->widthStep)
      m_pL1->m_pVirtualDepth[h]= (BYTE *) &(m_depthL1a->imageData[pos]);
    if(!m_pL1->depthsynthesis_reverse_3Dwarp(NULL)) return false;

    // C1位置に穴を生成しているので、シフトは無し

  } else { //////////////////////////////////////////////////////////////////////////
    if(!m_pS1->depthShift(m_pD1->m_pWorkDepth, from)) return false;
    cvSmooth(m_pS1->getVirtualDepthMap(), m_depthS1, CV_MEDIAN, 3);
    cvSmooth(m_pS1->getHolePixels(), m_imgMask[2], CV_MEDIAN, 3);
    depthInpaint(m_depthS1, m_imgMask[2]);

    for(h=pos=0; h<m_pL1->m_uiHeight; h++, pos+=m_depthS1->widthStep)
      m_pL1->m_pVirtualDepth[h]= (BYTE *) &(m_depthS1->imageData[pos]);
    if(!m_pL1->depthsynthesis_reverse_3Dwarp(NULL)) return false;

    // C1位置に穴を生成しているので、シフトは無し

  }
  printf("+");
//if (m_ResidualsSide == LEFT) sprintf(fname, "L1_d%03iL.png", fnum); else sprintf(fname, "L1_d%03iR.png", fnum);
//cvSaveImage(fname, m_pL1->getHolePixels());

  cvSet( m_imgTemp[0], cvScalarAll(255) );


  // A1:Step.1 L:V1 -> A1 -> C1, R:V2 -> A4 -> C3 ///////////////////////////////////
  if ( side == LEFT ) from= RIGHT; else from= LEFT;

  if(!m_pS2->depthShift(m_pD1->m_pWorkDepth, from)) return false;
  cvSmooth(m_pS2->getVirtualDepthMap(), m_depthS2, CV_MEDIAN, 3);
  cvSmooth(m_pS2->getHolePixels(), m_imgMask[3], CV_MEDIAN, 3);
  depthInpaint(m_depthS2, m_imgMask[3]);

  for(h=pos=0; h<m_pL2->m_uiHeight; h++, pos+=m_depthS2->widthStep)
    m_pL2->m_pVirtualDepth[h] = (BYTE *) &(m_depthS2->imageData[pos]);
  if(!m_pL2->depthsynthesis_reverse_3Dwarp(NULL)) return false;

  // C1へのシフト
  m_pS2d->setImgHoles(m_pL2->getHolePixels());
  if(!m_pS2d->depthShift(m_pL2->m_pVirtualDepth, from, HOLESHIFT)) return false;
  printf("+");
//if (m_ResidualsSide == LEFT) sprintf(fname, "S2_d%03iL.png", fnum); else sprintf(fname, "S2_d%03iR.png", fnum);
//cvSaveImage(fname, m_pS2d->getVirtualDepthMap());


  // A2:Step.1 L:V1 -> A2 -> C1, R:V2 -> A3 -> C3  //////////////////////////////////
  if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L )
    if ( side == LEFT ) from= RIGHT; else from= LEFT;
  else
    if ( side == LEFT ) from= LEFT; else from= RIGHT;
  if(!m_pS3->depthShift(m_pD1->m_pWorkDepth, from)) return false;
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_1shift_d%03iL.png", fnum); else sprintf(fname, "S3_1shift_d%03iR.png", fnum);
//cvSaveImage(fname, m_pS3->getVirtualDepthMap());
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_1shift_h%03iL.png", fnum); else sprintf(fname, "S3_1shift_h%03iR.png", fnum);
//cvSaveImage(fname, m_pS3->getHolePixels());
  cvSmooth(m_pS3->getVirtualDepthMap(), m_depthS3, CV_MEDIAN, 3);
  cvSmooth(m_pS3->getHolePixels(), m_imgMask[4], CV_MEDIAN, 3);
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_2mf_d%03iL.png", fnum); else sprintf(fname, "S3_2mf_d%03iR.png", fnum);
//cvSaveImage(fname, m_depthS3);
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_2mf_h%03iL.png", fnum); else sprintf(fname, "S3_2mf_h%03iR.png", fnum);
//cvSaveImage(fname, m_imgMask[4]);
  depthInpaint(m_depthS3, m_imgMask[4]);
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_3inpaint_d%03iL.png", fnum); else sprintf(fname, "S3_3inpaint_d%03iR.png", fnum);
//cvSaveImage(fname, m_depthS3);

  for(h=pos=0; h<m_pL3->m_uiHeight; h++, pos+=m_depthS3->widthStep)
    m_pL3->m_pVirtualDepth[h] = (BYTE *) &(m_depthS3->imageData[pos]);
  if(!m_pL3->depthsynthesis_reverse_3Dwarp(NULL)) return false;
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_4warp_h%03iL.png", fnum); else sprintf(fname, "S3_4warp_h%03iR.png", fnum);
//cvSaveImage(fname, m_pL3->getHolePixels());

  // C1へのシフト
  if ( side == LEFT ) from= RIGHT; else from= LEFT;
  m_pS3d->setImgHoles(m_pL3->getHolePixels());
  if(!m_pS3d->depthShift(m_pL3->m_pVirtualDepth, from, HOLESHIFT)) return false;
  printf("+");
//if (m_ResidualsSide == LEFT) sprintf(fname, "S3_d%03iL.png", fnum); else sprintf(fname, "S3_d%03iR.png", fnum);
//cvSaveImage(fname, m_pS3d->getVirtualDepthMap());
//return true;


  // V1:Step.2 L:V1 -> C2 -> C1, R:V2 -> C2 -> C3 ////////////////////////////////////
  //          Vartual Depth をMain側に投影した際の穴を取得
  if ( side == LEFT ) from= RIGHT; else from= LEFT;
  if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) { //////////////////////////////////////////////////////

    if(!m_pR1->depthShift(m_pD1->m_pWorkDepth, from)) return false;
    cvSmooth(m_pR1->getVirtualDepthMap(), m_depthR1, CV_MEDIAN, 3);
    cvSmooth(m_pR1->getHolePixels(), m_imgMask[3], CV_MEDIAN, 3);
    depthInpaint(m_depthR1, m_imgMask[3]);

    for(h=pos=0; h<m_pR1a->m_uiHeight; h++, pos+=m_depthR1->widthStep)
      m_pR1a->m_pVirtualDepth[h] = (BYTE *) &(m_depthR1->imageData[pos]);
    if(!m_pR1a->depthsynthesis_reverse_3Dwarp(NULL)) return false;

    // C1へのシフト
    m_pS1a->setImgHoles(m_pR1a->getHolePixels());
    if(!m_pS1a->depthShift(m_pR1a->m_pVirtualDepth, from, HOLESHIFT)) return false;

  } else { //////////////////////////////////////////////////////////////////////////

    m_pR1->setDepth(m_pD1->m_pWorkDepth);
    if(!m_pR1->depthsynthesis_reverse_3Dwarp(NULL)) return false;

    // C1へのシフト
    m_pS1->setImgHoles(m_pR1->getHolePixels());
    if(!m_pS1->depthShift(m_pR1->m_pVirtualDepth, from, HOLESHIFT)) return false;
    //cvSmooth(m_pS1->getVirtualDepthMap(), m_imgTemp[0], CV_MEDIAN, 3); // 穴！

  }
  printf("+");
//if (m_ResidualsSide == LEFT) sprintf(fname, "S1_d%03iL.png", fnum); else sprintf(fname, "S1_d%03iR.png", fnum);
//cvSaveImage(fname, m_pS1->getVirtualDepthMap());


  // Step.3  //////////////////////////////////////////////////
  //          取得した全穴L1/S2d/S3d/S1をマージし、Viewから残差を作る。
  //          S2d/S3d/S1の穴は、Depthに偽装されていることに注意。
  if ( mode == QUARTER_ODD || mode == QUARTER_ODD_L || mode == FULL_ODD_L ) { //////////////////////////////////////////////////////
    cvOr( m_pL1->getHolePixels(), m_pS2d->getVirtualDepthMap(), m_imgTemp[2] );
    cvOr( m_imgTemp[2], m_pS3d->getVirtualDepthMap(), m_imgTemp[3] );
    cvOr( m_imgTemp[3], m_pS1a->getVirtualDepthMap(), m_imgTemp[4] );
  } else { //////////////////////////////////////////////////////////////////////////
    cvOr( m_pL1->getHolePixels(), m_pS2d->getVirtualDepthMap(), m_imgTemp[2] );
    cvOr( m_imgTemp[2], m_pS3d->getVirtualDepthMap(), m_imgTemp[3] );
    cvOr( m_imgTemp[3], m_pS1->getVirtualDepthMap(), m_imgTemp[4] );
  }
  cvDilate(m_imgTemp[4], m_imgTemp[0], NULL, 8 ); // 4 -> 6 Jan17,2012 8 12 16 m_imgTemp[4]->width/120


//if (m_ResidualsSide == LEFT) sprintf(fname, "M_d%03iL.png", fnum); else sprintf(fname, "M_d%03iR.png", fnum);
//cvSaveImage(fname, m_imgTemp[0]);
//cvSaveImage(fname, m_imgTemp[0]);
//SynYuvBuffer->setDataFromDepth(m_imgTemp[0]);
//return true;

  if(!m_pL1->setView(CenterView)) return false;
  // CvScalar avg;
  if ( fnum % ONESET == 0 ) {
    cvSetImageCOI(m_pL1->getVirtualImage(), 1);
    avgColor.val[0]= cvAvg(m_pL1->getVirtualImage()).val[0];
    cvSetImageCOI(m_pL1->getVirtualImage(), 2);
    avgColor.val[1]= cvAvg(m_pL1->getVirtualImage()).val[0]; // 0
    cvSetImageCOI(m_pL1->getVirtualImage(), 3);
    avgColor.val[2]= cvAvg(m_pL1->getVirtualImage()).val[0]; // 0
    cvSetImageCOI(m_pL1->getVirtualImage(), 0);
  }
//if (m_ResidualsSide == LEFT)
//printf(" AvgL:%f,%f,%f, ", avg.val[0],avg.val[1],avg.val[2]);
//else
//printf(" AvgR:%f,%f,%f, ", avg.val[0],avg.val[1],avg.val[2]);

  // 投影画像は使わないので、素の位置の画像を読み直し
//  if( !m_pL1->setView(View) ) return false;
//  cvSet( m_imgInterpolatedView, avg ); // cvScalarAll(128));
//  cvCopy( m_pL1->getVirtualImage(), m_imgInterpolatedView, m_imgTemp[0] );
  if( !m_pL3->setView(View) ) return false;
  cvSet( m_imgInterpolatedView, avgColor ); // cvScalarAll(128));
  cvCopy( m_pL3->getVirtualImage(), m_imgInterpolatedView, m_imgTemp[0] );
  SynYuvBuffer->setTextureFromImgYUV444( m_imgInterpolatedView );
  AreaBuffer->setDepthFromImgYUV444( m_imgTemp[0] );

  return true;
}



// July6,2011 NiCT //////////////////////
void CResidualsInterpolation::ToFilterDepth(IplImage* pDepthSRC, IplImage* pDepthDST, IplImage* pDepthNEGSRC, IplImage* pDepthNEGDST, unsigned int fsize)
{
  // デプス穴マップSRCからデプス成功マップを作成
  // デプス穴マップSRCにMFかけてデプス穴マップDSTを作成
  // デプス成功マップとデプス穴マップDSTのANDをとってデプス有マスク作成
  // デプス有マスクをかけて、デプス穴マップSRCからデプス穴マップDSTにコピー
  cvNot(pDepthNEGSRC, m_depthSuccess);
  cvSmooth(pDepthNEGSRC, pDepthNEGDST, CV_MEDIAN, fsize);
  cvAnd(m_depthSuccess, pDepthNEGDST, m_imgMask[4]);
  cvCopy(pDepthNEGSRC, pDepthNEGDST, m_imgMask[4]);

  // デプスSRCに」MDかけてデプスDSTを作成
  // デプス有マスクをかけて、デプスSRCからデプスDSTにコピー
  cvSmooth(pDepthSRC, pDepthDST, CV_MEDIAN, fsize);
  cvCopy(pDepthSRC, pDepthDST, m_imgMask[4]);
}
